import argparse
import asyncio
import logging

from amqtt.client import MQTTClient
from amqtt.mqtt.constants import QOS_1, QOS_2

"""
This sample shows how to publish messages to secure broker.

Use `openssl req -x509 -nodes -days 365 -newkey rsa:2048 -keyout key.pem -out cert.pem -subj "/CN=localhost"` to
generate a self-signed certificate for the broker to use.
"""

logger = logging.getLogger(__name__)

config = {
    "will": {
        "topic": "/will/client",
        "message": "Dead or alive",
        "qos": QOS_1,
        "retain": True,
    },
    "auto_reconnect": False,
    "check_hostname": False,
    "certfile": "",
}


async def test_coro(certfile: str) -> None:
    config["certfile"] = certfile
    client = MQTTClient(config=config)

    await client.connect("mqtts://localhost:8883")
    tasks = [
        asyncio.ensure_future(client.publish("a/b", b"TEST MESSAGE WITH QOS_0")),
        asyncio.ensure_future(client.publish("a/b", b"TEST MESSAGE WITH QOS_1", qos=QOS_1)),
        asyncio.ensure_future(client.publish("a/b", b"TEST MESSAGE WITH QOS_2", qos=QOS_2)),
    ]
    await asyncio.wait(tasks)
    logger.info("messages published")
    await client.disconnect()


def __main__():
    formatter = "[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.DEBUG, format=formatter)
    parser = argparse.ArgumentParser()
    parser.add_argument("--cert", default="cert.pem", help="path & file to verify server's authenticity")
    args = parser.parse_args()

    asyncio.run(test_coro(args.cert))

if __name__ == "__main__":
    __main__()



import asyncio
import logging
import unittest

from amqtt.events import BrokerEvents

from amqtt.plugins.base import BaseAuthPlugin, BaseTopicPlugin
from amqtt.plugins.manager import PluginManager
from amqtt.contexts import BaseContext, Action
from amqtt.session import Session

formatter = "[%(asctime)s] %(name)s {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
logging.basicConfig(level=logging.INFO, format=formatter)


class EmptyTestPlugin:
    def __init__(self, context: BaseContext) -> None:
        self.context = context


class EventTestPlugin(BaseAuthPlugin, BaseTopicPlugin):
    def __init__(self, context: BaseContext) -> None:
        super().__init__(context)
        self.test_close_flag = False
        self.test_auth_flag = False
        self.test_topic_flag = False
        self.test_event_flag = False

    async def on_broker_message_received(self) -> None:
        self.test_event_flag = True

    async def authenticate(self, *, session: Session) -> bool | None:
        self.test_auth_flag = True
        return True

    async def topic_filtering(
        self, *, session: Session | None = None, topic: str | None = None, action: Action | None = None
    ) -> bool:
        self.test_topic_flag = True
        return False

    async def close(self) -> None:
        self.test_close_flag = True


class TestPluginManager(unittest.TestCase):
    def setUp(self) -> None:
        self.loop = asyncio.new_event_loop()

    def test_load_plugin(self) -> None:
        manager = PluginManager("amqtt.test.plugins", context=None)
        assert len(manager._plugins) > 0

    def test_fire_event(self) -> None:
        async def fire_event() -> None:
            await manager.fire_event(BrokerEvents.MESSAGE_RECEIVED)
            await asyncio.sleep(1)
            await manager.close()

        manager = PluginManager("amqtt.test.plugins", context=None)
        self.loop.run_until_complete(fire_event())
        plugin = manager.get_plugin("EventTestPlugin")
        assert plugin is not None
        assert plugin.test_event_flag

    def test_fire_event_wait(self) -> None:
        async def fire_event() -> None:
            await manager.fire_event(BrokerEvents.MESSAGE_RECEIVED, wait=True)
            await manager.close()

        manager = PluginManager("amqtt.test.plugins", context=None)
        self.loop.run_until_complete(fire_event())
        plugin = manager.get_plugin("EventTestPlugin")
        assert plugin is not None
        assert plugin.test_event_flag

    def test_plugin_close_coro(self) -> None:

        manager = PluginManager("amqtt.test.plugins", context=None)
        self.loop.run_until_complete(manager.map_plugin_close())
        self.loop.run_until_complete(asyncio.sleep(0.5))
        plugin = manager.get_plugin("EventTestPlugin")
        assert plugin is not None
        assert plugin.test_close_flag

    def test_plugin_auth_coro(self) -> None:
        # provide context that activates auth plugins
        context = BaseContext()
        context.config = {'auth':{}}

        manager = PluginManager("amqtt.test.plugins", context=context)
        self.loop.run_until_complete(manager.map_plugin_auth(session=Session()))
        self.loop.run_until_complete(asyncio.sleep(0.5))
        plugin = manager.get_plugin("EventTestPlugin")
        assert plugin is not None
        assert plugin.test_auth_flag

    def test_plugin_topic_coro(self) -> None:
        # provide context that activates topic check plugins
        context = BaseContext()
        context.config = {'topic-check':{}}

        manager = PluginManager("amqtt.test.plugins", context=context)
        self.loop.run_until_complete(manager.map_plugin_topic(session=Session(), topic="test", action=Action.PUBLISH))
        self.loop.run_until_complete(asyncio.sleep(0.5))
        plugin = manager.get_plugin("EventTestPlugin")
        assert plugin is not None
        assert plugin.test_topic_flag



import asyncio
from dataclasses import dataclass
import logging

from amqtt.broker import Broker
from amqtt.plugins.base import BasePlugin
from amqtt.session import Session

"""
This sample shows how to run a broker without stacktraces on keyboard interrupt
"""

logger = logging.getLogger(__name__)


class RemoteInfoPlugin(BasePlugin):

    async def on_broker_client_connected(self, *, client_id:str, client_session:Session) -> None:
        display_port_str = f"on port '{client_session.remote_port}'" if self.config.display_port else ""

        logger.info(f"client '{client_id}' connected from"
                    f" '{client_session.remote_address}' {display_port_str}")

    @dataclass
    class Config:
        display_port: bool = False

config = {
    "listeners": {
        "default": {
            "type": "tcp",
            "bind": "0.0.0.0:1883",
        },
        "ws-mqtt": {
            "bind": "127.0.0.1:8080",
            "type": "ws",
            "max_connections": 10,
        },
    },
    "plugins": {
        "amqtt.plugins.authentication.AnonymousAuthPlugin": { "allow_anonymous": True},
        "samples.broker_custom_plugin.RemoteInfoPlugin": { "display_port": True },
    }
}

async def main_loop():
    broker = Broker(config)
    try:
        await broker.start()
        while True:
            await asyncio.sleep(1)
    except asyncio.CancelledError:
        await broker.shutdown()

async def main():
    t = asyncio.create_task(main_loop())
    try:
        await t
    except asyncio.CancelledError:
        pass

def __main__():

    formatter = "[%(asctime)s] :: %(levelname)s :: %(name)s :: %(message)s"
    logging.basicConfig(level=logging.INFO, format=formatter)

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    task = loop.create_task(main())

    try:
        loop.run_until_complete(task)
    except KeyboardInterrupt:
        logger.info("KeyboardInterrupt received. Stopping server...")
        task.cancel()
        loop.run_until_complete(task)  # Ensure task finishes cleanup
    finally:
        logger.info("Server stopped.")
        loop.close()

if __name__ == "__main__":
    __main__()



import asyncio
import logging

from amqtt.client import ClientError, MQTTClient
from amqtt.mqtt.constants import QOS_1, QOS_2

"""
This sample shows how to subscribe to different $SYS topics and how to receive incoming messages 
"""

logger = logging.getLogger(__name__)


async def uptime_coro() -> None:
    client = MQTTClient(config={"auto_reconnect": False})
    await client.connect("mqtt://localhost:1883")

    await client.subscribe(
        [
            ("$SYS/broker/uptime", QOS_1),
            ("$SYS/broker/load/#", QOS_2),
        ],
    )
    logger.info("Subscribed")
    try:
        for _i in range(1, 10):
            if msg := await client.deliver_message():
                logger.info(f"{msg.topic} >> {msg.data.decode()}")
        await client.unsubscribe(["$SYS/broker/uptime", "$SYS/broker/load/#"])
        logger.info("UnSubscribed")
        await client.disconnect()
    except ClientError:
        logger.exception("Client exception")


def __main__():
    formatter = "[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=formatter)
    asyncio.run(uptime_coro())

if __name__ == "__main__":
    __main__()



import asyncio
import unittest

import pytest

from amqtt.adapters import BufferReader
from amqtt.errors import MQTTError
from amqtt.mqtt.packet import CONNECT, MQTTFixedHeader


class TestMQTTFixedHeaderTest(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()

    def test_from_bytes(self):
        data = b"\x10\x7f"
        stream = BufferReader(data)
        header = self.loop.run_until_complete(MQTTFixedHeader.from_stream(stream))
        assert header.packet_type == CONNECT
        assert not header.flags & 8
        assert (header.flags & 6) >> 1 == 0
        assert not header.flags & 1
        assert header.remaining_length == 127

    def test_from_bytes_with_length(self):
        data = b"\x10\xff\xff\xff\x7f"
        stream = BufferReader(data)
        header = self.loop.run_until_complete(MQTTFixedHeader.from_stream(stream))
        assert header.packet_type == CONNECT
        assert not header.flags & 8
        assert (header.flags & 6) >> 1 == 0
        assert not header.flags & 1
        assert header.remaining_length == 268435455

    def test_from_bytes_ko_with_length(self):
        data = b"\x10\xff\xff\xff\xff\x7f"
        stream = BufferReader(data)
        with pytest.raises(MQTTError):
            self.loop.run_until_complete(MQTTFixedHeader.from_stream(stream))

    def test_to_bytes(self):
        header = MQTTFixedHeader(CONNECT, 0x00, 0)
        data = header.to_bytes()
        assert data == b"\x10\x00"

    def test_to_bytes_2(self):
        header = MQTTFixedHeader(CONNECT, 0x00, 268435455)
        data = header.to_bytes()
        assert data == b"\x10\xff\xff\xff\x7f"



import asyncio
import logging

from amqtt.client import ClientError, MQTTClient
from amqtt.mqtt.constants import QOS_1

"""
Run `samples/broker_acl.py` or `samples/broker_taboo.py` 

This sample shows how to subscribe to different topics, some of which are allowed. 
"""

logger = logging.getLogger(__name__)


async def uptime_coro() -> None:
    client = MQTTClient()
    await client.connect("mqtt://test:test@0.0.0.0:1883")

    result = await client.subscribe(
        [
            ("$SYS/#", QOS_1),  # Topic forbidden when running `broker_acl.py`
            ("data/memes", QOS_1),  # Topic allowed
            ("data/classified", QOS_1),  # Topic forbidden
            ("repositories/amqtt/master", QOS_1),  # Topic allowed
            ("repositories/amqtt/devel", QOS_1),  # Topic forbidden when running `broker_acl.py`
            ("calendar/amqtt/releases", QOS_1),  # Topic allowed
        ],
    )
    logger.info(f"Subscribed results: {result}")
    try:
        for _i in range(1, 100):
            if msg := await client.deliver_message():
                logger.info(f"{msg.topic} >> {msg.data.decode()}")
        await client.unsubscribe(["$SYS/#", "data/memes"])
        logger.info("UnSubscribed")
        await client.disconnect()
    except ClientError:
        logger.exception("Client exception")


def __main__():
    formatter = "[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=formatter)
    asyncio.get_event_loop().run_until_complete(uptime_coro())

if __name__ == "__main__":
    __main__()



import asyncio
import logging
from pathlib import Path

import typer
from yaml.parser import ParserError

from amqtt import __version__ as amqtt_version
from amqtt.broker import Broker
from amqtt.errors import BrokerError, PluginError
from amqtt.utils import read_yaml_config

logger = logging.getLogger(__name__)


app = typer.Typer(add_completion=False, rich_markup_mode=None)


def main() -> None:
    """Run the MQTT broker."""
    app()


def _version(v: bool) -> None:
    if v:
        typer.echo(f"{amqtt_version}")
        raise typer.Exit(code=0)


@app.command()
def broker_main(
        config_file: str | None = typer.Option(None, "-c", help="broker configuration file"),
        debug: bool = typer.Option(False, "-d", help="Enable debug messages"),
        version: bool = typer.Option(  # noqa : ARG001
            False,
            "--version",
            callback=_version,
            is_eager=True,
            help="Show version and exit",
        ),
) -> None:
    """Command-line script for running a MQTT 3.1.1 broker."""
    formatter = "[%(asctime)s] :: %(levelname)s - %(message)s"
    if debug:
        formatter = "[%(asctime)s] %(name)s:%(lineno)d :: %(levelname)s - %(message)s"

    level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(level=level, format=formatter)
    logging.getLogger("transitions").setLevel(logging.WARNING)
    try:
        if config_file:
            config = read_yaml_config(config_file)
        else:
            config = read_yaml_config(Path(__file__).parent / "default_broker.yaml")
            logger.debug("Using default configuration")
    except FileNotFoundError as exc:
        typer.echo(f"❌ Config file error: {exc}", err=True)
        raise typer.Exit(code=1) from exc

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        broker = Broker(config, loop=loop)
    except (BrokerError, ParserError, PluginError) as exc:
        typer.echo(f"❌ Broker failed to start: {exc}", err=True)
        raise typer.Exit(code=1) from exc

    _ = loop.create_task(broker.start())  # noqa : RUF006
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        loop.run_until_complete(broker.shutdown())
    except Exception as exc:
        typer.echo("❌ Broker execution halted", err=True)
        raise typer.Exit(code=1) from exc
    finally:
        loop.close()


if __name__ == "__main__":
    main()



import asyncio
import logging
import random
import threading
import time
from pathlib import Path
from threading import Thread
from typing import Any
from unittest.mock import MagicMock, call, patch

import pytest
import yaml
from paho.mqtt import client as paho_client
from yaml import Loader

from amqtt.broker import Broker
from amqtt.events import BrokerEvents
from amqtt.client import MQTTClient
from amqtt.mqtt.constants import QOS_1, QOS_2
from amqtt.session import Session

logger = logging.getLogger(__name__)
paho_logger = logging.getLogger("paho_client")


# monkey patch MagicMock
# taken from https://stackoverflow.com/questions/51394411/python-object-magicmock-cant-be-used-in-await-expression
async def async_magic():
    pass


MagicMock.__await__ = lambda _: async_magic().__await__()

@pytest.mark.asyncio
async def test_paho_connect(broker, mock_plugin_manager):

    test_complete = asyncio.Event()

    host = "localhost"
    port = 1883
    client_id = f'python-mqtt-{random.randint(0, 1000)}'

    def on_connect(client, userdata, flags, rc, properties=None):
        assert rc == 0, f"Connection failed with result code {rc}"
        client.disconnect()

    def on_disconnect(client, userdata, flags, rc, properties=None):
        assert rc == 0, f"Disconnect failed with result code {rc}"
        test_complete.set()

    test_client = paho_client.Client(paho_client.CallbackAPIVersion.VERSION2, client_id=client_id)
    test_client.enable_logger(paho_logger)

    test_client.on_connect = on_connect
    test_client.on_disconnect = on_disconnect

    test_client.connect(host, port)
    test_client.loop_start()

    await asyncio.wait_for(test_complete.wait(), timeout=5)
    await asyncio.sleep(0.1)

    broker.plugins_manager.fire_event.assert_called()
    assert broker.plugins_manager.fire_event.call_count > 2

    # double indexing is ugly, but call_args_list returns a tuple of tuples
    events = [c[0][0] for c in broker.plugins_manager.fire_event.call_args_list]
    assert BrokerEvents.CLIENT_CONNECTED in events
    assert BrokerEvents.CLIENT_DISCONNECTED in events

    test_client.loop_stop()


@pytest.mark.asyncio
async def test_paho_qos1(broker, mock_plugin_manager):

    sub_client = MQTTClient()
    await sub_client.connect("mqtt://127.0.0.1")
    ret = await sub_client.subscribe(
        [("/qos1", QOS_1),],
    )

    host = "localhost"
    port = 1883
    client_id = f'python-mqtt-{random.randint(0, 1000)}'

    test_client = paho_client.Client(paho_client.CallbackAPIVersion.VERSION2, client_id=client_id)
    test_client.enable_logger(paho_logger)

    test_client.connect(host, port)
    test_client.loop_start()
    await asyncio.sleep(0.1)
    test_client.publish("/qos1", "test message", qos=1)
    await asyncio.sleep(0.1)
    test_client.loop_stop()

    message = await sub_client.deliver_message()
    assert message is not None
    assert message.qos == 1
    assert message.topic == "/qos1"
    assert message.data == b"test message"
    await sub_client.disconnect()
    await asyncio.sleep(0.1)


@pytest.mark.asyncio
async def test_paho_qos2(broker, mock_plugin_manager):
    sub_client = MQTTClient()
    await sub_client.connect("mqtt://127.0.0.1")
    ret = await sub_client.subscribe(
        [("/qos2", QOS_2), ],
    )

    host = "localhost"
    port = 1883
    client_id = f'python-mqtt-{random.randint(0, 1000)}'

    test_client = paho_client.Client(paho_client.CallbackAPIVersion.VERSION2, client_id=client_id)
    test_client.enable_logger(paho_logger)

    test_client.connect(host, port)
    test_client.loop_start()
    await asyncio.sleep(0.1)
    test_client.publish("/qos2", "test message", qos=2)
    await asyncio.sleep(0.1)
    test_client.loop_stop()

    message = await sub_client.deliver_message()
    assert message is not None
    assert message.qos == 2
    assert message.topic == "/qos2"
    assert message.data == b"test message"
    await sub_client.disconnect()
    await asyncio.sleep(0.1)



def run_paho_client(flag):
    client_id = 'websocket_client_1'
    logging.info("creating paho client")
    test_client = paho_client.Client(callback_api_version=paho_client.CallbackAPIVersion.VERSION2,
                                     transport='websockets',
                                     client_id=client_id)

    test_client.ws_set_options('')
    logging.info("client connecting...")
    test_client.connect('127.0.0.1', 8080)
    logging.info("starting loop")
    test_client.loop_start()
    logging.info("client connected")
    time.sleep(1)
    logging.info("sending messages")
    test_client.publish("/qos2", "test message", qos=2)
    test_client.publish("/qos2", "test message", qos=2)
    test_client.publish("/qos2", "test message", qos=2)
    test_client.publish("/qos2", "test message", qos=2)
    time.sleep(1)
    test_client.loop_stop()
    test_client.disconnect()
    flag.set()


@pytest.mark.asyncio
async def test_paho_ws():
    path = Path('docs_test/test.amqtt.local.yaml')
    with path.open() as f:
        cfg: dict[str, Any] = yaml.load(f, Loader=Loader)
    logger.warning(cfg)
    broker = Broker(config=cfg)
    await broker.start()

    # python websockets and paho mqtt don't play well with each other in the same thread
    flag = threading.Event()
    thread = Thread(target=run_paho_client, args=(flag,))
    thread.start()

    await asyncio.sleep(5)
    thread.join(1)

    assert flag.is_set(), "paho thread didn't execute completely"

    logging.info("client disconnected")
    await broker.shutdown()



import logging
from typing import Any

try:
    from enum import Enum, StrEnum
except ImportError:
    # support for python 3.10
    from enum import Enum
    class StrEnum(str, Enum):  #type: ignore[no-redef]
        pass

from dacite import from_dict, Config

from amqtt.contexts import BrokerConfig, ListenerType

logger = logging.getLogger(__name__)


def test_entrypoint_broker_config(caplog):
    test_cfg: dict[str, Any] = {
        "listeners": {
            "default": {"type": "tcp", "bind": "127.0.0.1:1883", "max_connections": 10},
        },
        'sys_interval': 1,
        'auth': {
            'allow_anonymous': True
        }
    }
    if 'plugins' not in test_cfg:
        test_cfg['plugins'] = None
    # cfg: dict[str, Any] = yaml.load(config, Loader=Loader)


    broker_config = from_dict(data_class=BrokerConfig, data=test_cfg, config=Config(cast=[StrEnum, ListenerType]))
    assert isinstance(broker_config, BrokerConfig)

    assert broker_config.plugins is None





from typing_extensions import Self

from amqtt.errors import AMQTTError
from amqtt.mqtt.packet import UNSUBACK, MQTTFixedHeader, MQTTPacket, PacketIdVariableHeader


class UnsubackPacket(MQTTPacket[PacketIdVariableHeader, None, MQTTFixedHeader]):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = None

    def __init__(
        self,
        fixed: MQTTFixedHeader | None = None,
        variable_header: PacketIdVariableHeader | None = None,
        payload: None = None,
    ) -> None:
        if fixed is None:
            header = MQTTFixedHeader(UNSUBACK, 0x00)
        else:
            if fixed.packet_type is not UNSUBACK:
                msg = f"Invalid fixed packet type {fixed.packet_type} for UnsubackPacket init"
                raise AMQTTError(
                    msg,
                )
            header = fixed

        super().__init__(header)
        self.variable_header = variable_header
        self.payload = payload

    @classmethod
    def build(cls, packet_id: int) -> Self:
        variable_header = PacketIdVariableHeader(packet_id)
        return cls(variable_header=variable_header)



import logging
from pathlib import Path
import sys

import typer

logger = logging.getLogger(__name__)

app = typer.Typer(add_completion=False, rich_markup_mode=None)


def main() -> None:
    """Run the `server_creds` cli."""
    app()


@app.command()
def server_creds(
        country: str = typer.Option(..., "--country", help="x509 'country_name' attribute"),
        org_name: str = typer.Option(..., "--org-name", help="x509 'organization_name' attribute"),
        cn: str = typer.Option(..., "--cn", help="x509 'common_name' attribute"),
        output_dir: str = typer.Option(Path.cwd().absolute(), "--output-dir", help="output directory"),
        ca_key_fn: str = typer.Option("ca.key", "--ca-key", help="server key output filename."),
        ca_crt_fn: str = typer.Option("ca.crt", "--ca-crt", help="server cert output filename."),
) -> None:
    """Generate a key and certificate for the broker in pem format, signed by the provided CA credentials. With a key size of 2048 and a 1-year expiration."""  # noqa : E501
    formatter = "[%(asctime)s] :: %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=formatter)
    try:
        from amqtt.contrib.cert import (  # pylint: disable=import-outside-toplevel
            generate_server_csr,
            load_ca,
            sign_csr,
            write_key_and_crt,
        )
    except ImportError:
        msg = "Requires installation of the optional 'contrib' package: `pip install amqtt[contrib]`"
        logger.critical(msg)
        sys.exit(1)

    ca_key, ca_crt = load_ca(ca_key_fn, ca_crt_fn)

    server_key, server_csr = generate_server_csr(country=country, org_name=org_name, cn=cn)
    server_crt = sign_csr(server_csr, ca_key, ca_crt)

    write_key_and_crt(server_key, server_crt, "server", Path(output_dir))


if __name__ == "__main__":
    main()



import asyncio
import logging
from pathlib import Path

from amqtt.broker import Broker

"""
This sample shows how to run a broker with the topic check taboo plugin
"""

logger = logging.getLogger(__name__)

config = {
    "listeners": {
        "default": {
            "type": "tcp",
            "bind": "0.0.0.0:1883",
        },
        "ws-mqtt": {
            "bind": "127.0.0.1:8080",
            "type": "ws",
            "max_connections": 10,
        },
    },
    "plugins": {
        "amqtt.plugins.authentication.AnonymousAuthPlugin": {"allow_anonymous": True},
        "amqtt.plugins.authentication.FileAuthPlugin": {
            "password_file": Path(__file__).parent / "passwd",
        },
        "amqtt.plugins.sys.broker.BrokerSysPlugin": {"sys_interval": 10},
        "amqtt.plugins.topic_checking.TopicTabooPlugin": {},
    }
}


async def main_loop():
    broker = Broker(config)
    try:
        await broker.start()
        while True:
            await asyncio.sleep(1)
    except asyncio.CancelledError:
        await broker.shutdown()


async def main():
    t = asyncio.create_task(main_loop())
    try:
        await t
    except asyncio.CancelledError:
        pass


def __main__():

    formatter = "[%(asctime)s] :: %(levelname)s :: %(name)s :: %(message)s"
    logging.basicConfig(level=logging.INFO, format=formatter)

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    task = loop.create_task(main())

    try:
        loop.run_until_complete(task)
    except KeyboardInterrupt:
        logger.info("KeyboardInterrupt received. Stopping server...")
        task.cancel()
        loop.run_until_complete(task)  # Ensure task finishes cleanup
    finally:
        logger.info("Server stopped.")
        loop.close()

if __name__ == "__main__":
    __main__()



from collections.abc import Sequence
from dataclasses import asdict
import logging
import time
from typing import Any, Optional
import uuid

from sqlalchemy import JSON, CheckConstraint, Integer, String, UniqueConstraint, desc, event, func, select
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncSession
from sqlalchemy.orm import DeclarativeBase, Mapped, Mapper, Session, make_transient, mapped_column

from amqtt.contrib.shadows.states import StateDocument

logger = logging.getLogger(__name__)


class ShadowUpdateError(Exception):
    def __init__(self, message: str = "updating an existing Shadow is not allowed") -> None:
        super().__init__(message)


class ShadowBase(DeclarativeBase):
    pass


async def sync_shadow_base(connection: AsyncConnection) -> None:
    """Create tables and table schemas."""
    await connection.run_sync(ShadowBase.metadata.create_all)


def default_state_document() -> dict[str, Any]:
    """Create a default (empty) state document, factory for model field."""
    return asdict(StateDocument())


class Shadow(ShadowBase):
    __tablename__ = "shadows_shadow"

    id: Mapped[str | None] = mapped_column(String(36), primary_key=True, default=lambda: str(uuid.uuid4()))

    device_id: Mapped[str] = mapped_column(String(128), nullable=False)
    name: Mapped[str] = mapped_column(String(128), nullable=False)
    version: Mapped[int] = mapped_column(Integer, nullable=False)

    _state: Mapped[dict[str, Any]] = mapped_column("state", JSON, nullable=False, default=dict)

    created_at: Mapped[int] = mapped_column(Integer, default=lambda: int(time.time()), nullable=False)

    __table_args__ = (
        CheckConstraint("version > 0", name="check_quantity_positive"),
        UniqueConstraint("device_id", "name", "version", name="uq_device_id_name_version"),
    )

    @property
    def state(self) -> StateDocument:
        if not self._state:
            return StateDocument()
        return StateDocument.from_dict(self._state)

    @state.setter
    def state(self, value: StateDocument) -> None:
        self._state = asdict(value)

    @classmethod
    async def latest_version(cls, session: AsyncSession, device_id: str, name: str) -> Optional["Shadow"]:
        """Get the latest version of the shadow associated with the device and name."""
        stmt = (
            select(cls).where(
                cls.device_id == device_id,
                cls.name == name
            ).order_by(desc(cls.version)).limit(1)
        )
        result = await session.execute(stmt)
        return result.scalar_one_or_none()

    @classmethod
    async def all(cls, session: AsyncSession, device_id: str, name: str) -> Sequence["Shadow"]:
        """Return a list of all shadows associated with the device and name."""
        stmt = (
                select(cls).where(
                    cls.device_id == device_id,
                    cls.name == name
                ).order_by(desc(cls.version)))
        result = await session.execute(stmt)
        return result.scalars().all()


@event.listens_for(Shadow, "before_insert")
def assign_incremental_version(_: Mapper[Any], connection: Session, target: "Shadow") -> None:
    """Get the latest version of the state document."""
    stmt = (
        select(func.max(Shadow.version))
        .where(
            Shadow.device_id == target.device_id,
            Shadow.name == target.name
        )
    )
    result = connection.execute(stmt).scalar_one_or_none()
    target.version = (result or 0) + 1


@event.listens_for(Shadow, "before_update")
def prevent_update(_mapper: Mapper[Any], _session: Session, _instance: "Shadow") -> None:
    """Prevent existing shadow from being updated."""
    raise ShadowUpdateError


@event.listens_for(Session, "before_flush")
def convert_update_to_insert(session: Session, _flush_context: object, _instances: object | None) -> None:
    """Force a shadow to insert a new version, instead of updating an existing."""
    # Make a copy of the dirty set so we can safely mutate the session
    dirty = list(session.dirty)

    for obj in dirty:
        if not session.is_modified(obj, include_collections=False):
            continue  # skip unchanged

        # You can scope this to a particular class
        if not isinstance(obj, Shadow):
            continue

        # Clone logic: convert update into insert
        session.expunge(obj)         # remove from session
        make_transient(obj)          # remove identity and history
        obj.id = ""                  # clear primary key
        obj.version += 1             # bump version or modify fields

        session.add(obj)             # re-add as new object


_listener_example = '''#
# @event.listens_for(Shadow, "before_insert")
# def convert_state_document_to_json(_1: Mapper[Any], _2: Session, target: "Shadow") -> None:
#     """Listen for insertion and convert state document to json."""
#     if not isinstance(target.state, StateDocument):
#         msg = "'state' field needs to be a StateDocument"
#         raise TypeError(msg)
#
#     target.state = target.state.to_dict()
'''



"""INIT."""

__all__ = ["MQTTPacket"]

from typing import Any, TypeAlias

from amqtt.errors import AMQTTError
from amqtt.mqtt.connack import ConnackPacket
from amqtt.mqtt.connect import ConnectPacket
from amqtt.mqtt.disconnect import DisconnectPacket
from amqtt.mqtt.packet import (
    CONNACK,
    CONNECT,
    DISCONNECT,
    PINGREQ,
    PINGRESP,
    PUBACK,
    PUBCOMP,
    PUBLISH,
    PUBREC,
    PUBREL,
    SUBACK,
    SUBSCRIBE,
    UNSUBACK,
    UNSUBSCRIBE,
    MQTTFixedHeader,
    MQTTPacket,
)
from amqtt.mqtt.pingreq import PingReqPacket
from amqtt.mqtt.pingresp import PingRespPacket
from amqtt.mqtt.puback import PubackPacket
from amqtt.mqtt.pubcomp import PubcompPacket
from amqtt.mqtt.publish import PublishPacket
from amqtt.mqtt.pubrec import PubrecPacket
from amqtt.mqtt.pubrel import PubrelPacket
from amqtt.mqtt.suback import SubackPacket
from amqtt.mqtt.subscribe import SubscribePacket
from amqtt.mqtt.unsuback import UnsubackPacket
from amqtt.mqtt.unsubscribe import UnsubscribePacket

_P: TypeAlias = MQTTPacket[Any, Any, Any]

packet_dict: dict[int, type[_P]] = {
    CONNECT: ConnectPacket,
    CONNACK: ConnackPacket,
    PUBLISH: PublishPacket,
    PUBACK: PubackPacket,
    PUBREC: PubrecPacket,
    PUBREL: PubrelPacket,
    PUBCOMP: PubcompPacket,
    SUBSCRIBE: SubscribePacket,
    SUBACK: SubackPacket,
    UNSUBSCRIBE: UnsubscribePacket,
    UNSUBACK: UnsubackPacket,
    PINGREQ: PingReqPacket,
    PINGRESP: PingRespPacket,
    DISCONNECT: DisconnectPacket,
}


def packet_class(fixed_header: MQTTFixedHeader) -> type[_P]:
    """Return the packet class for a given fixed header.

    :param fixed_header: The fixed header of the packet.
    :type
        fixed_header: MQTTFixedHeader
    :return: The packet class for the given fixed header.
    :rtype: type[MQTTPacket]
    :raises AMQTTError: If the packet type is not recognized.
    """
    if fixed_header.packet_type not in packet_dict:
        msg = f"Unexpected packet Type '{fixed_header.packet_type}'"
        raise AMQTTError(msg)
    try:
        return packet_dict[fixed_header.packet_type]
    except KeyError as e:
        msg = f"Unexpected packet Type '{fixed_header.packet_type}'"
        raise AMQTTError(msg) from e



import logging
import subprocess
from pathlib import Path
import tempfile
from typing import Any
import unittest.mock
import urllib.request

import pytest

from amqtt.broker import Broker
from amqtt.contexts import BaseContext
from amqtt.plugins.base import BasePlugin

log = logging.getLogger(__name__)

pytest_plugins = ["pytest_logdog"]

@pytest.fixture
def rsa_keys():
    tmp_dir = tempfile.TemporaryDirectory(prefix='amqtt-test-')
    cert = Path(tmp_dir.name) / "cert.pem"
    key = Path(tmp_dir.name) / "key.pem"
    cmd = f'openssl req -x509 -nodes -days 365 -newkey rsa:2048 -keyout {key} -out {cert} -subj "/CN=localhost"'
    subprocess.run(cmd, shell=True, capture_output=True, text=True)
    yield cert, key
    tmp_dir.cleanup()


@pytest.fixture
def test_config(rsa_keys):
    certfile, keyfile = rsa_keys
    yield {
        "listeners": {
            "default": {"type": "tcp", "bind": "127.0.0.1:1883", "max_connections": 15},
            "mqtts": {
                "type": "tcp",
                "bind": "127.0.0.1:1884",
                "max_connections": 15,
                "ssl": True,
                "certfile": certfile,
                "keyfile": keyfile
            },
            "ws": {"type": "ws", "bind": "127.0.0.1:8080", "max_connections": 15},
            "wss": {
                "type": "ws",
                "bind": "127.0.0.1:8081",
                "max_connections": 15,
                "ssl": True,
                'certfile': certfile,
                'keyfile': keyfile},
        },
        "sys_interval": 0,
        "auth": {
            "allow-anonymous": True,
        }
    }

test_config_acl: dict[str, int | dict[str, Any]] = {
    "listeners": {
        "default": {"type": "tcp", "bind": "127.0.0.1:1884", "max_connections": 10},
    },
    "sys_interval": 0,
    "auth": {
        "plugins": ["auth_file"],
        "password-file": Path(__file__).resolve().parent / "plugins" / "passwd",
    },
    "topic-check": {
        "enabled": True,
        "plugins": ["topic_acl"],
        "acl": {
            "user1": ["public/#"],
            "user2": ["#"],
        },
        "publish-acl": {"user1": ["public/subtopic/#"]},
    },
}


@pytest.fixture
def mock_plugin_manager():
    with (unittest.mock.patch("amqtt.broker.PluginManager") as plugin_manager):
        plugin_manager_instance = plugin_manager.return_value

        # disable topic filtering when using the mock manager
        plugin_manager_instance.is_topic_filtering_enabled.return_value = False

        # allow any connection when using the mock manager
        plugin_manager_instance.map_plugin_auth = unittest.mock.AsyncMock(return_value={ BasePlugin(BaseContext()): True })

        yield plugin_manager


@pytest.fixture
async def broker_fixture(test_config):
    with pytest.warns(DeprecationWarning):
        broker = Broker(test_config, plugin_namespace="amqtt.test.plugins")
        await broker.start()
        assert broker.transitions.is_started()
        assert broker._sessions == {}
        assert "default" in broker._servers

        yield broker

    if not broker.transitions.is_stopped():
        await broker.shutdown()


@pytest.fixture
async def broker(mock_plugin_manager, test_config):
    # just making sure the mock is in place before we start our broker
    assert mock_plugin_manager is not None

    broker = Broker(test_config, plugin_namespace="amqtt.test.plugins")
    await broker.start()
    assert broker.transitions.is_started()
    assert broker._sessions == {}
    assert "default" in broker._servers

    yield broker

    if not broker.transitions.is_stopped():
        await broker.shutdown()


@pytest.fixture
async def acl_broker():
    broker = Broker(
        test_config_acl,
        plugin_namespace="amqtt.broker.plugins",
    )
    await broker.start()
    assert broker.transitions.is_started()
    assert broker._sessions == {}
    assert "default" in broker._servers

    yield broker

    if not broker.transitions.is_stopped():
        await broker.shutdown()


@pytest.fixture(scope="module")
def ca_file_fixture():
    temp_dir = Path(tempfile.mkdtemp(prefix="amqtt-test-"))
    url = "http://test.mosquitto.org/ssl/mosquitto.org.crt"
    ca_file = temp_dir / "mosquitto.org.crt"
    urllib.request.urlretrieve(url, str(ca_file))
    log.info(f"Stored mosquitto cert at {ca_file}")

    # Yield the CA file path for tests
    yield ca_file

    # Cleanup after the tests
    if temp_dir.exists():
        for file in temp_dir.iterdir():
            file.unlink()
        temp_dir.rmdir()


def pytest_addoption(parser):
    parser.addoption(
        "--mock-docker",
        action="store",
        default="false",
        help="for environments where docker isn't available, mock calls which require docker",
    )


"""Plugin to determine authentication of clients with DB storage."""
from dataclasses import dataclass

import click

try:
    from enum import StrEnum
except ImportError:
    # support for python 3.10
    from enum import Enum
    class StrEnum(str, Enum):  # type: ignore[no-redef]
        pass

from .plugin import TopicAuthDBPlugin, UserAuthDBPlugin


class DBType(StrEnum):
    """Enumeration for supported relational databases."""

    MARIA = "mariadb"
    MYSQL = "mysql"
    POSTGRESQL = "postgresql"
    SQLITE = "sqlite"


@dataclass
class DBInfo:
    """SQLAlchemy database information."""

    connect_str: str
    connect_port: int | None


_db_map = {
    DBType.MARIA: DBInfo("mysql+aiomysql", 3306),
    DBType.MYSQL: DBInfo("mysql+aiomysql", 3306),
    DBType.POSTGRESQL: DBInfo("postgresql+asyncpg", 5432),
    DBType.SQLITE: DBInfo("sqlite+aiosqlite", None)
}


def db_connection_str(db_type: DBType, db_username: str, db_host: str, db_port: int | None, db_filename: str) -> str:
    """Create sqlalchemy database connection string."""
    db_info = _db_map[db_type]
    if db_type == DBType.SQLITE:
        return f"{db_info.connect_str}:///{db_filename}"
    db_password = click.prompt("Enter the db password (press enter for none)", hide_input=True)
    pwd = f":{db_password}" if db_password else ""
    return f"{db_info.connect_str}://{db_username}:{pwd}@{db_host}:{db_port or db_info.connect_port}"


__all__ = ["DBType", "TopicAuthDBPlugin", "UserAuthDBPlugin", "db_connection_str"]



import asyncio
import logging
import shutil
import subprocess
import tempfile
from pathlib import Path

from unittest.mock import MagicMock

from OpenSSL import crypto

import pytest

from amqtt.broker import BrokerContext, Broker
from amqtt.client import MQTTClient
from amqtt.contrib.cert import UserAuthCertPlugin
from amqtt.errors import ConnectError
from amqtt.scripts.server_creds import server_creds as get_server_creds
from amqtt.scripts.device_creds import device_creds as get_device_creds
from amqtt.scripts.ca_creds import ca_creds as get_ca_creds
from amqtt.session import Session

logger = logging.getLogger(__name__)

@pytest.fixture
def temp_directory():
    temp_dir = Path(tempfile.mkdtemp(prefix="amqtt-test-"))
    yield temp_dir
    logger.critical(temp_dir)
    # shutil.rmtree(temp_dir)

@pytest.fixture
def ca_creds(temp_directory):

    get_ca_creds(country='US', state="NY", locality="NYC", org_name="aMQTT", cn="aMQTT", output_dir=str(temp_directory))

    ca_key = temp_directory / "ca.key"
    ca_crt = temp_directory / "ca.crt"
    return ca_key, ca_crt

@pytest.fixture
def server_creds(ca_creds, temp_directory):
    ca_key = temp_directory / "ca.key"
    ca_crt = temp_directory / "ca.crt"
    get_server_creds(country='US', org_name='aMQTT', cn='aMQTT',
                     output_dir=str(temp_directory),
                     ca_key_fn=str(ca_key), ca_crt_fn=str(ca_crt))
    server_key = temp_directory / "server.key"
    server_crt = temp_directory / "server.crt"
    yield server_key, server_crt

@pytest.fixture
def device_creds(ca_creds, temp_directory):
    ca_key, ca_crt = ca_creds
    get_device_creds(country='US', org_name='aMQTT',
                     device_id="mydeviceid", uri='test.amqtt.io',
                     output_dir=str(temp_directory),
                     ca_key_fn=str(ca_key), ca_crt_fn=str(ca_crt))
    yield temp_directory / "mydeviceid.key", temp_directory / "mydeviceid.crt"

def test_device_cert(temp_directory, ca_creds, server_creds, device_creds):
    ca_key, ca_crt = ca_creds
    server_key, server_crt = server_creds
    device_key, device_crt = device_creds

    assert ca_key.exists()
    assert ca_crt.exists()
    assert server_key.exists()
    assert server_crt.exists()
    assert device_key.exists()
    assert device_crt.exists()

    r = subprocess.run(f"openssl x509 -in {str(device_crt)} -noout -text", shell=True, capture_output=True, text=True, check=True)

    assert "URI:spiffe://test.amqtt.io/device/mydeviceid, DNS:mydeviceid.local" in r.stdout

@pytest.fixture
def ssl_object_mock(device_creds):
    device_key, device_crt = device_creds

    with device_crt.open("rb") as f:
        cert = crypto.load_certificate(crypto.FILETYPE_PEM, f.read())
    mock = MagicMock()
    mock.getpeercert.return_value = crypto.dump_certificate(crypto.FILETYPE_ASN1, cert)
    yield mock


@pytest.mark.parametrize("uri_domain,client_id,expected_result", [
    ('test.amqtt.io', 'mydeviceid', True),
    ('test.amqtt.io', 'notmydeviceid', False),
    ('other.amqtt.io', 'mydeviceid', False),
])
@pytest.mark.asyncio
async def test_cert_plugin(ssl_object_mock, uri_domain, client_id, expected_result):

    empty_cfg = {
        'listeners': {'default': {'type':'tcp', 'bind':'127.0.0.1:1883'}},
        'plugins': {}
    }

    bc = BrokerContext(broker=Broker(config=empty_cfg))
    bc.config = UserAuthCertPlugin.Config(uri_domain=uri_domain)

    cert_auth_plugin = UserAuthCertPlugin(bc)

    s = Session()
    s.client_id = client_id
    s.ssl_object = ssl_object_mock

    assert await cert_auth_plugin.authenticate(session=s) == expected_result


@pytest.mark.asyncio
async def test_client_broker_cert_authentication(ca_creds, server_creds, device_creds):
    ca_key, ca_crt = ca_creds
    server_key, server_crt = server_creds
    device_key, device_crt = device_creds
    broker_config = {
        'listeners': {
            'default': {
                'type':'tcp',
                'bind':'127.0.0.1:8883',
                'ssl': True,
                'keyfile': server_key,
                'certfile': server_crt,
                'cafile': ca_crt,
            }
        },
        'plugins': {
            'amqtt.plugins.logging_amqtt.PacketLoggerPlugin':{},
            'amqtt.contrib.cert.UserAuthCertPlugin': {'uri_domain': 'test.amqtt.io'},
        }
    }

    b = Broker(config=broker_config)
    await b.start()
    await asyncio.sleep(1)

    client_config = {
        'auto_reconnect': False,
        'broker': {
            'cafile': ca_crt,
            'certfile': device_crt,
            'keyfile': device_key
        }
    }

    c = MQTTClient(config=client_config, client_id='mydeviceid')
    await c.connect('mqtts://127.0.0.1:8883')
    await asyncio.sleep(0.1)

    assert 'mydeviceid' in b._sessions
    s, _ = b._sessions['mydeviceid']
    assert s.transitions.state == "connected"

    await asyncio.sleep(0.1)
    await c.disconnect()
    await asyncio.sleep(0.1)
    await b.shutdown()


def ssl_error_logger(loop, context):
    logger.critical("Asyncio SSL error:", context.get("message"))
    exc = repr(context.get("exception"))
    assert "exception" not in context, f"Exception: {exc}"


@pytest.mark.asyncio
async def test_client_broker_wrong_certs(ca_creds, server_creds, device_creds):
    loop = asyncio.get_event_loop()
    loop.set_exception_handler(ssl_error_logger)
    loop.set_debug(True)

    ca_key, ca_crt = ca_creds
    server_key, server_crt = server_creds
    device_key, device_crt = device_creds
    broker_config = {
        'listeners': {
            'default': {
                'type':'tcp',
                'bind':'127.0.0.1:8883',
                'ssl': True,
                'keyfile': server_key,
                'certfile': server_crt,
                'cafile': ca_crt,
            }
        },
        'plugins': {
            'amqtt.plugins.logging_amqtt.PacketLoggerPlugin':{},
            'amqtt.contrib.cert.UserAuthCertPlugin': {'uri_domain': 'test.amqtt.io'},
        }
    }

    b = Broker(config=broker_config)
    await b.start()
    await asyncio.sleep(1)

    # generate a different ca certificate and make sure the connection fails
    temp_dir = Path(tempfile.mkdtemp(prefix="amqtt-test-"))
    get_ca_creds(country='US', state="NY", locality="NYC", org_name="aMQTT", cn="aMQTT", output_dir=str(temp_dir))
    wrong_ca_crt = temp_dir / 'ca.crt'
    client_config = {
        'auto_reconnect': False,
        'connection': {
            'cafile': wrong_ca_crt,
            'certfile': device_crt,
            'keyfile': device_key,
        }
    }

    c = MQTTClient(config=client_config, client_id='mydeviceid')
    with pytest.raises(ConnectError, match='.+?SSL: CERTIFICATE_VERIFY_FAILED.+?'):
        await c.connect('mqtts://127.0.0.1:8883')

    await b.shutdown()


from dataclasses import dataclass
import logging
from typing import ClassVar

import jwt

try:
    from enum import StrEnum
except ImportError:
    # support for python 3.10
    from enum import Enum
    class StrEnum(str, Enum):  # type: ignore[no-redef]
        pass

from amqtt.broker import BrokerContext
from amqtt.contexts import Action
from amqtt.plugins import TopicMatcher
from amqtt.plugins.base import BaseAuthPlugin, BaseTopicPlugin
from amqtt.session import Session

logger = logging.getLogger(__name__)


class Algorithms(StrEnum):
    ES256 = "ES256"
    ES256K = "ES256K"
    ES384 = "ES384"
    ES512 = "ES512"
    ES521 = "ES521"
    EdDSA = "EdDSA"
    HS256 = "HS256"
    HS384 = "HS384"
    HS512 = "HS512"
    PS256 = "PS256"
    PS384 = "PS384"
    PS512 = "PS512"
    RS256 = "RS256"
    RS384 = "RS384"
    RS512 = "RS512"


class UserAuthJwtPlugin(BaseAuthPlugin):

    async def authenticate(self, *, session: Session) -> bool | None:

        if not session.username or not session.password:
            return None

        try:
            decoded_payload = jwt.decode(session.password, self.config.secret_key, algorithms=["HS256"])
            return bool(decoded_payload.get(self.config.user_claim, None) == session.username)
        except jwt.ExpiredSignatureError:
            logger.debug(f"jwt for '{session.username}' is expired")
            return False
        except jwt.InvalidTokenError:
            logger.debug(f"jwt for '{session.username}' is invalid")
            return False

    @dataclass
    class Config:
        """Configuration for the JWT user authentication."""

        secret_key: str
        """Secret key to decrypt the token."""
        user_claim: str
        """Payload key for user name."""
        algorithm: str = "HS256"
        """Algorithm to use for token encryption: 'ES256', 'ES256K', 'ES384', 'ES512', 'ES521', 'EdDSA', 'HS256',
         'HS384', 'HS512', 'PS256', 'PS384', 'PS512', 'RS256', 'RS384', 'RS512'"""


class TopicAuthJwtPlugin(BaseTopicPlugin):

    _topic_jwt_claims: ClassVar = {
        Action.PUBLISH: "publish_claim",
        Action.SUBSCRIBE: "subscribe_claim",
        Action.RECEIVE: "receive_claim",
    }

    def __init__(self, context: BrokerContext) -> None:
        super().__init__(context)

        self.topic_matcher = TopicMatcher()

    async def topic_filtering(
        self, *, session: Session | None = None, topic: str | None = None, action: Action | None = None
    ) -> bool | None:

        if not session or not topic or not action:
            return None

        if not session.password:
            return None

        try:
            decoded_payload = jwt.decode(session.password.encode(), self.config.secret_key, algorithms=["HS256"])
            claim = getattr(self.config, self._topic_jwt_claims[action])
            return any(self.topic_matcher.is_topic_allowed(topic, a_filter) for a_filter in decoded_payload.get(claim, []))
        except jwt.ExpiredSignatureError:
            logger.debug(f"jwt for '{session.username}' is expired")
            return False
        except jwt.InvalidTokenError:
            logger.debug(f"jwt for '{session.username}' is invalid")
            return False

    @dataclass
    class Config:
        """Configuration for the JWT topic authorization."""

        secret_key: str
        """Secret key to decrypt the token."""
        publish_claim: str
        """Payload key for contains a list of permissible publish topics."""
        subscribe_claim: str
        """Payload key for contains a list of permissible subscribe topics."""
        receive_claim: str
        """Payload key for contains a list of permissible receive topics."""
        algorithm: str = "HS256"
        """Algorithm to use for token encryption: 'ES256', 'ES256K', 'ES384', 'ES512', 'ES521', 'EdDSA', 'HS256',
         'HS384', 'HS512', 'PS256', 'PS384', 'PS512', 'RS256', 'RS384', 'RS512'"""



import asyncio
import logging
import secrets

try:
    from datetime import UTC, datetime, timedelta
except ImportError:
    from datetime import datetime, timezone, timedelta

    UTC = timezone.utc

import jwt
import pytest

from amqtt.broker import BrokerContext, Broker
from amqtt.client import MQTTClient
from amqtt.contexts import Action, ListenerConfig, BrokerConfig
from amqtt.contrib.jwt import UserAuthJwtPlugin, TopicAuthJwtPlugin
from amqtt.mqtt.constants import QOS_0
from amqtt.session import Session


@pytest.fixture
def secret_key():
    return secrets.token_urlsafe(32)


@pytest.mark.parametrize("exp_time, outcome", [
    (datetime.now(UTC) + timedelta(hours=1), True),
    (datetime.now(UTC) - timedelta(hours=1), False),
])
@pytest.mark.asyncio
async def test_user_jwt_plugin(secret_key, exp_time, outcome):

    payload = {
        "username": "example_user",
        "exp": exp_time
    }

    ctx = BrokerContext(Broker())
    ctx.config = UserAuthJwtPlugin.Config(
        secret_key=secret_key,
        user_claim='username'
    )

    jwt_plugin = UserAuthJwtPlugin(context=ctx)

    s = Session()
    s.username = "example_user"
    s.password = jwt.encode(payload, secret_key, algorithm="HS256")

    assert await jwt_plugin.authenticate(session=s) == outcome, "access should have been granted"


@pytest.mark.asyncio
async def test_topic_jwt_plugin(secret_key):

    payload = {
        "username": "example_user",
        "exp": datetime.now(UTC) + timedelta(hours=1),
        "publish_acl": ['my/topic/#', 'my/+/other']
    }

    ctx = BrokerContext(Broker())
    ctx.config = TopicAuthJwtPlugin.Config(
        secret_key=secret_key,
        publish_claim='publish_acl',
        subscribe_claim='subscribe_acl',
        receive_claim='receive_acl'
    )

    jwt_plugin = TopicAuthJwtPlugin(context=ctx)

    s = Session()
    s.username = "example_user"
    s.password = jwt.encode(payload, secret_key, algorithm="HS256")

    assert await jwt_plugin.topic_filtering(session=s, topic="my/topic/one", action=Action.PUBLISH), "access should be granted"


@pytest.mark.asyncio
async def test_broker_with_jwt_plugin(secret_key, caplog):
    payload = {
        "username": "example_user",
        "exp": datetime.now(UTC) + timedelta(hours=1),
        "publish_acl": ['my/topic/#', 'my/+/other'],
        "subscribe_acl": ['my/+/other'],
    }
    username = "example_user"
    password = jwt.encode(payload, secret_key, algorithm="HS256")

    cfg = BrokerConfig(
        listeners={'default': ListenerConfig()},
        plugins={
            'amqtt.contrib.jwt.UserAuthJwtPlugin': {
                'secret_key': secret_key,
                'user_claim': 'username',
            },
            'amqtt.contrib.jwt.TopicAuthJwtPlugin': {
                'secret_key': secret_key,
                'publish_claim': 'publish_acl',
                'subscribe_claim': 'subscribe_acl',
                'receive_claim': 'receive_acl'
            }
        }
    )
    with caplog.at_level(logging.INFO):
        b = Broker(config=cfg)
        await b.start()
        await asyncio.sleep(0.1)

        c = MQTTClient()
        await c.connect(f'mqtt://{username}:{password}@localhost:1883')
        await asyncio.sleep(0.1)
        result = await c.subscribe([('my/one', QOS_0)])
        assert result == [128, ]
        result = await c.subscribe([('my/one/other', QOS_0)])
        assert result == [0]
        await c.publish('my/one', b'message should not get published')
        await asyncio.sleep(0.1)
        assert "not allowed to publish to TOPIC my/one" in caplog.text
        await asyncio.sleep(0.1)

        await c.disconnect()
        await asyncio.sleep(0.1)
        await b.shutdown()



from asyncio import StreamReader
from typing_extensions import Self

from amqtt.adapters import ReaderAdapter
from amqtt.codecs_amqtt import (
    bytes_to_int,
    decode_data_with_length,
    decode_string,
    encode_data_with_length,
    encode_string,
    int_to_bytes,
    read_or_raise,
)
from amqtt.errors import AMQTTError, NoDataError
from amqtt.mqtt.packet import CONNECT, MQTTFixedHeader, MQTTPacket, MQTTPayload, MQTTVariableHeader
from amqtt.utils import gen_client_id


class ConnectVariableHeader(MQTTVariableHeader):
    __slots__ = ("flags", "keep_alive", "proto_level", "proto_name")

    USERNAME_FLAG = 0x80
    PASSWORD_FLAG = 0x40
    WILL_RETAIN_FLAG = 0x20
    WILL_FLAG = 0x04
    WILL_QOS_MASK = 0x18
    CLEAN_SESSION_FLAG = 0x02
    RESERVED_FLAG = 0x01

    def __init__(self, connect_flags: int = 0x00, keep_alive: int = 0, proto_name: str = "MQTT", proto_level: int = 0x04) -> None:
        super().__init__()
        self.proto_name = proto_name
        self.proto_level = proto_level
        self.flags = connect_flags
        self.keep_alive = keep_alive

    def __repr__(self) -> str:
        """Return a string representation of the ConnectVariableHeader object."""
        return (
            f"ConnectVariableHeader(proto_name={self.proto_name}, proto_level={self.proto_level},"
            f" flags={hex(self.flags)}, keepalive={self.keep_alive})"
        )

    def _set_flag(self, val: bool, mask: int) -> None:
        if val:
            self.flags |= mask
        else:
            self.flags &= ~mask

    def _get_flag(self, mask: int) -> bool:
        return bool(self.flags & mask)

    @classmethod
    async def from_stream(cls, reader: ReaderAdapter, _: MQTTFixedHeader) -> Self:
        #  protocol name
        protocol_name = await decode_string(reader)

        # protocol level
        protocol_level_byte = await read_or_raise(reader, 1)
        protocol_level = bytes_to_int(protocol_level_byte)

        # flags
        flags_byte = await read_or_raise(reader, 1)
        flags = bytes_to_int(flags_byte)

        # keep-alive
        keep_alive_byte = await read_or_raise(reader, 2)
        keep_alive = bytes_to_int(keep_alive_byte)

        return cls(flags, keep_alive, protocol_name, protocol_level)

    def to_bytes(self) -> bytes | bytearray:
        out = bytearray()

        # Protocol name
        out.extend(encode_string(self.proto_name))
        # Protocol level
        out.append(self.proto_level)
        # flags
        out.append(self.flags)
        # keep alive
        out.extend(int_to_bytes(self.keep_alive, 2))

        return out

    @property
    def username_flag(self) -> bool:
        return self._get_flag(self.USERNAME_FLAG)

    @username_flag.setter
    def username_flag(self, val: bool) -> None:
        self._set_flag(val, self.USERNAME_FLAG)

    @property
    def password_flag(self) -> bool:
        return self._get_flag(self.PASSWORD_FLAG)

    @password_flag.setter
    def password_flag(self, val: bool) -> None:
        self._set_flag(val, self.PASSWORD_FLAG)

    @property
    def will_retain_flag(self) -> bool:
        return self._get_flag(self.WILL_RETAIN_FLAG)

    @will_retain_flag.setter
    def will_retain_flag(self, val: bool) -> None:
        self._set_flag(val, self.WILL_RETAIN_FLAG)

    @property
    def will_flag(self) -> bool:
        return self._get_flag(self.WILL_FLAG)

    @will_flag.setter
    def will_flag(self, val: bool) -> None:
        self._set_flag(val, self.WILL_FLAG)

    @property
    def clean_session_flag(self) -> bool:
        return self._get_flag(self.CLEAN_SESSION_FLAG)

    @clean_session_flag.setter
    def clean_session_flag(self, val: bool) -> None:
        self._set_flag(val, self.CLEAN_SESSION_FLAG)

    @property
    def reserved_flag(self) -> bool:
        return self._get_flag(self.RESERVED_FLAG)

    @reserved_flag.setter
    def reserved_flag(self, val: bool) -> None:
        self._set_flag(val, self.RESERVED_FLAG)

    @property
    def will_qos(self) -> int:
        return (self.flags & 0x18) >> 3

    @will_qos.setter
    def will_qos(self, val: int) -> None:
        self.flags &= 0xE7  # Reset QOS flags
        self.flags |= val << 3


class ConnectPayload(MQTTPayload[ConnectVariableHeader]):
    __slots__ = (
        "client_id",
        "client_id_is_random",
        "password",
        "username",
        "will_message",
        "will_topic",
    )

    def __init__(
        self,
        client_id: str | None = None,
        will_topic: str | None = None,
        will_message: bytes | bytearray | None = None,
        username: str | None = None,
        password: str | None = None,
    ) -> None:
        super().__init__()
        self.client_id_is_random = False
        self.client_id = client_id
        self.will_topic = will_topic
        self.will_message = will_message
        self.username = username
        self.password = password

    def __repr__(self) -> str:
        """Return a string representation of the ConnectPayload object."""
        return (
            f"ConnectVariableHeader(client_id={self.client_id}, will_topic={self.will_topic},"
            f"will_message={self.will_message!r}, username={self.username}, password={self.password})"
        )

    @classmethod
    async def from_stream(
        cls,
        reader: StreamReader | ReaderAdapter,
        _: MQTTFixedHeader | None,
        variable_header: ConnectVariableHeader | None,
    ) -> Self:
        payload = cls()
        #  Client identifier
        try:
            payload.client_id = await decode_string(reader)
        except NoDataError:
            payload.client_id = None

        if payload.client_id is None or payload.client_id == "":
            # A Server MAY allow a Client to supply a ClientId that has a length of zero bytes
            # [MQTT-3.1.3-6]
            payload.client_id = gen_client_id()
            # indicator to throw exception in case CLEAN_SESSION_FLAG is set to False
            payload.client_id_is_random = True

        # Read will topic, username and password
        if variable_header is not None and variable_header.will_flag:
            try:
                payload.will_topic = await decode_string(reader)
                payload.will_message = await decode_data_with_length(reader)
            except NoDataError:
                payload.will_topic = None
                payload.will_message = None

        if variable_header is not None and variable_header.username_flag:
            try:
                payload.username = await decode_string(reader)
            except NoDataError:
                payload.username = None

        if variable_header is not None and variable_header.password_flag:
            try:
                payload.password = await decode_string(reader)
            except NoDataError:
                payload.password = None

        return payload

    def to_bytes(
        self,
        fixed_header: MQTTFixedHeader | None = None,
        variable_header: ConnectVariableHeader | None = None,
    ) -> bytes | bytearray:
        out = bytearray()
        # Client identifier
        if self.client_id is not None:
            out.extend(encode_string(self.client_id))
        # Will topic / message
        if variable_header is not None and variable_header.will_flag:
            if self.will_topic is not None:
                out.extend(encode_string(self.will_topic))
            if self.will_message is not None:
                out.extend(encode_data_with_length(self.will_message))
        # username
        if variable_header is not None and variable_header.username_flag and self.username is not None:
            out.extend(encode_string(self.username))
        # password
        if variable_header is not None and variable_header.password_flag and self.password is not None:
            out.extend(encode_string(self.password))

        return out


class ConnectPacket(MQTTPacket[ConnectVariableHeader, ConnectPayload, MQTTFixedHeader]):  # type: ignore [type-var]
    VARIABLE_HEADER = ConnectVariableHeader
    PAYLOAD = ConnectPayload

    def __init__(
        self,
        fixed: MQTTFixedHeader | None = None,
        variable_header: ConnectVariableHeader | None = None,
        payload: ConnectPayload | None = None,
    ) -> None:
        if fixed is None:
            header = MQTTFixedHeader(CONNECT, 0x00)
        else:
            if fixed.packet_type is not CONNECT:
                msg = f"Invalid fixed packet type {fixed.packet_type} for ConnectPacket init"
                raise AMQTTError(msg)
            header = fixed
        super().__init__(header)
        self.variable_header = variable_header
        self.payload = payload

    @property
    def proto_name(self) -> str:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        return self.variable_header.proto_name

    @proto_name.setter
    def proto_name(self, name: str) -> None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        self.variable_header.proto_name = name

    @property
    def proto_level(self) -> int:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        return self.variable_header.proto_level

    @proto_level.setter
    def proto_level(self, level: int) -> None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        self.variable_header.proto_level = level

    @property
    def username_flag(self) -> bool:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        return self.variable_header.username_flag

    @username_flag.setter
    def username_flag(self, flag: bool) -> None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        self.variable_header.username_flag = flag

    @property
    def password_flag(self) -> bool:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        return self.variable_header.password_flag

    @password_flag.setter
    def password_flag(self, flag: bool) -> None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        self.variable_header.password_flag = flag

    @property
    def clean_session_flag(self) -> bool:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        return self.variable_header.clean_session_flag

    @clean_session_flag.setter
    def clean_session_flag(self, flag: bool) -> None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        self.variable_header.clean_session_flag = flag

    @property
    def will_retain_flag(self) -> bool:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        return self.variable_header.will_retain_flag

    @will_retain_flag.setter
    def will_retain_flag(self, flag: bool) -> None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        self.variable_header.will_retain_flag = flag

    @property
    def will_qos(self) -> int:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        return self.variable_header.will_qos

    @will_qos.setter
    def will_qos(self, flag: int) -> None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        self.variable_header.will_qos = flag

    @property
    def will_flag(self) -> bool:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        return self.variable_header.will_flag

    @will_flag.setter
    def will_flag(self, flag: bool) -> None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        self.variable_header.will_flag = flag

    @property
    def reserved_flag(self) -> bool:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        return self.variable_header.reserved_flag

    @reserved_flag.setter
    def reserved_flag(self, flag: bool) -> None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        self.variable_header.reserved_flag = flag

    @property
    def client_id(self) -> str | None:
        if self.payload is None:
            msg = "Payload is not set"
            raise ValueError(msg)
        return self.payload.client_id

    @client_id.setter
    def client_id(self, client_id: str) -> None:
        if self.payload is None:
            msg = "Payload is not set"
            raise ValueError(msg)
        self.payload.client_id = client_id

    @property
    def client_id_is_random(self) -> bool:
        if self.payload is None:
            msg = "Payload is not set"
            raise ValueError(msg)
        return self.payload.client_id_is_random

    @client_id_is_random.setter
    def client_id_is_random(self, client_id_is_random: bool) -> None:
        if self.payload is None:
            msg = "Payload is not set"
            raise ValueError(msg)
        self.payload.client_id_is_random = client_id_is_random

    @property
    def will_topic(self) -> str | None:
        if self.payload is None:
            msg = "Payload is not set"
            raise ValueError(msg)
        return self.payload.will_topic

    @will_topic.setter
    def will_topic(self, will_topic: str) -> None:
        if self.payload is None:
            msg = "Payload is not set"
            raise ValueError(msg)
        self.payload.will_topic = will_topic

    @property
    def will_message(self) -> bytes | bytearray | None:
        if self.payload is None:
            msg = "Payload is not set"
            raise ValueError(msg)
        return self.payload.will_message

    @will_message.setter
    def will_message(self, will_message: bytes | bytearray) -> None:
        if self.payload is None:
            msg = "Payload is not set"
            raise ValueError(msg)
        self.payload.will_message = will_message

    @property
    def username(self) -> str | None:
        if self.payload is None:
            msg = "Payload is not set"
            raise ValueError(msg)
        return self.payload.username

    @username.setter
    def username(self, username: str) -> None:
        if self.payload is None:
            msg = "Payload is not set"
            raise ValueError(msg)
        self.payload.username = username

    @property
    def password(self) -> str | None:
        if self.payload is None:
            msg = "Payload is not set"
            raise ValueError(msg)
        return self.payload.password

    @password.setter
    def password(self, password: str) -> None:
        if self.payload is None:
            msg = "Payload is not set"
            raise ValueError(msg)
        self.payload.password = password

    @property
    def keep_alive(self) -> int:
        if self.variable_header is None:
            msg = "Payload is not set"
            raise ValueError(msg)
        return self.variable_header.keep_alive

    @keep_alive.setter
    def keep_alive(self, keep_alive: int) -> None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        self.variable_header.keep_alive = keep_alive



get_schema = {
  "$schema": "http://json-schema.org/draft-07/schema#",
  "title": "AWS IoT Shadow Get Request",
  "type": "object",
  "properties": {
    "clientToken": { "type": "string" }
  },
  "additionalProperties": False
}

get_accepted_schema = {
  "$schema": "http://json-schema.org/draft-07/schema#",
  "title": "AWS IoT Shadow Get Accepted",
  "type": "object",
  "properties": {
    "state": {
      "type": "object",
      "properties": {
        "desired": { "type": "object" },
        "reported": { "type": "object" }
      }
    },
    "metadata": {
      "type": "object"
    },
    "version": { "type": "integer" },
    "timestamp": { "type": "integer" },
    "clientToken": { "type": "string" }
  },
  "required": ["state", "version", "timestamp"],
  "additionalProperties": False
}

get_rejected_schema = {
  "$schema": "http://json-schema.org/draft-07/schema#",
  "title": "AWS IoT Shadow Get Rejected",
  "type": "object",
  "properties": {
    "code": { "type": "integer" },
    "message": { "type": "string" },
    "clientToken": { "type": "string" }
  },
  "required": ["code", "message"],
  "additionalProperties": False
}

update_schema = {
  "$schema": "http://json-schema.org/draft-07/schema#",
  "title": "AWS IoT Shadow Update",
  "type": "object",
  "properties": {
    "state": {
      "type": "object",
      "properties": {
        "desired": { "type": "object" },
        "reported": { "type": "object" }
      },
      "additionalProperties": False
    },
    "clientToken": { "type": "string" },
    "version": { "type": "integer" }
  },
  "additionalProperties": False
}

update_accepted_schema = {
  "$schema": "http://json-schema.org/draft-07/schema#",
  "title": "AWS IoT Shadow Update Accepted",
  "type": "object",
  "properties": {
    "state": {
      "type": "object",
      "properties": {
        "desired": { "type": "object" },
        "reported": { "type": "object" }
      }
    },
    "metadata": {
      "type": "object",
      "properties": {
        "desired": { "type": "object" },
        "reported": { "type": "object" }
      }
    },
    "version": { "type": "integer" },
    "timestamp": { "type": "integer" },
    "clientToken": { "type": "string" }
  },
  "required": ["version", "timestamp"],
  "additionalProperties": False
}

update_rejected_schema = {
  "$schema": "http://json-schema.org/draft-07/schema#",
  "title": "AWS IoT Shadow Update Rejected",
  "type": "object",
  "properties": {
    "code": { "type": "integer" },
    "message": { "type": "string" },
    "clientToken": { "type": "string" }
  },
  "required": ["code", "message"],
  "additionalProperties": False
}

delta_schema = {
  "$schema": "http://json-schema.org/draft-07/schema#",
  "title": "AWS IoT Shadow Delta",
  "type": "object",
  "properties": {
    "state": { "type": "object" },
    "metadata": { "type": "object" },
    "version": { "type": "integer" },
    "timestamp": { "type": "integer" }
  },
  "required": ["state", "version", "timestamp"],
  "additionalProperties": False
}

update_documents_schema = {
  "$schema": "http://json-schema.org/draft-07/schema#",
  "title": "AWS IoT Shadow Update Documents",
  "type": "object",
  "properties": {
    "previous": {
      "type": "object",
      "properties": {
        "state": {
          "type": "object",
          "properties": {
            "desired": { "type": "object" },
            "reported": { "type": "object" }
          },
          "additionalProperties": True
        },
        "metadata": {
          "type": "object",
          "properties": {
            "desired": { "type": "object" },
            "reported": { "type": "object" }
          },
          "additionalProperties": True
        },
        "version": { "type": "integer" },
        "timestamp": { "type": "integer" }
      },
      "required": ["state", "metadata", "version", "timestamp"],
      "additionalProperties": False
    },
    "current": {
      "type": "object",
      "properties": {
        "state": {
          "type": "object",
          "properties": {
            "desired": { "type": "object" },
            "reported": { "type": "object" }
          },
          "additionalProperties": True
        },
        "metadata": {
          "type": "object",
          "properties": {
            "desired": { "type": "object" },
            "reported": { "type": "object" }
          },
          "additionalProperties": True
        },
        "version": { "type": "integer" },
        "timestamp": { "type": "integer" }
      },
      "required": ["state", "metadata", "version", "timestamp"],
      "additionalProperties": False
    },
    "timestamp": { "type": "integer" }
  },
  "required": ["previous", "current", "timestamp"],
  "additionalProperties": False
}



import ast
import pprint
from typing import Any

import griffe
from griffe import Inspector, ObjectNode, Visitor, Attribute

from amqtt.contexts import default_listeners, default_broker_plugins, default_client_plugins
from amqtt.contrib.auth_db.plugin import default_hash_scheme

default_factory_map = {
    'default_listeners': default_listeners(),
    'default_broker_plugins': default_broker_plugins(),
    'default_client_plugins': default_client_plugins(),
    'default_hash_scheme': default_hash_scheme()
}

def get_qualified_name(node: ast.AST) -> str | None:
    """Recursively build the qualified name from an AST node."""
    if isinstance(node, ast.Name):
        return node.id
    elif isinstance(node, ast.Attribute):
        parent = get_qualified_name(node.value)
        if parent:
            return f"{parent}.{node.attr}"
        return node.attr
    elif isinstance(node, ast.Call):
        # e.g., uuid.uuid4()
        return get_qualified_name(node.func)
    return None

def get_fully_qualified_name(call_node):
    """
    Extracts the fully qualified name from an ast.Call node.
    """
    if isinstance(call_node.func, ast.Name):
        # Direct function call (e.g., "my_function(arg)")
        return call_node.func.id
    elif isinstance(call_node.func, ast.Attribute):
        # Method call or qualified name (e.g., "obj.method(arg)" or "module.submodule.function(arg)")
        parts = []
        current = call_node.func
        while isinstance(current, ast.Attribute):
            parts.append(current.attr)
            current = current.value
        if isinstance(current, ast.Name):
            parts.append(current.id)
        return ".".join(reversed(parts))
    else:
        # Handle other potential cases (e.g., ast.Subscript) if necessary
        return None

def get_callable_name(node):
    if isinstance(node, ast.Name):
        return node.id
    elif isinstance(node, ast.Attribute):
        return f"{get_callable_name(node.value)}.{node.attr}"
    return None

def evaluate_callable_node(node):
    try:
        # Wrap the node in an Expression so it can be compiled
        expr = ast.Expression(body=node)
        compiled = compile(expr, filename="<ast>", mode="eval")
        return eval(compiled, {"__builtins__": __builtins__, "list": list, "dict": dict})
    except Exception as e:
        return f"<unresolvable: {e}>"

class DataclassDefaultFactoryExtension(griffe.Extension):
    """Renders the output of a dataclasses field which uses a default factory.

    def other_field_defaults():
        return {'item1': 'value1', 'item2': 'value2'}

    @dataclass
    class MyDataClass:
        my_field: dict[str, Any] = field(default_factory=dict)
        my_other_field: dict[str, Any] = field(default_factory=other_field_defaults)

    instead of documentation rendering this as:

    ```
      class MyDataClass:
        my_field: dict[str, Any] = dict()
        my_other_field: dict[str, Any] = other_field_defaults()
    ```

    it will be displayed with the output of factory functions for more clarity:

    ```
    class MyDataClass:
        my_field: dict[str, Any] = {}
        my_other_field: dict[str, Any] = {'item1': 'value1', 'item2': 'value2'}
    ```

    _note_ : for any custom default factory function, it must be added to the `default_factory_map`
    in this file as `griffe` doesn't provide a straightforward mechanism with its AST to dynamically
    import/call the function.
    """

    def on_attribute_instance(
        self,
        *,
        node: ast.AST | ObjectNode,
        attr: Attribute,
        agent: Visitor | Inspector,
        **kwargs: Any,
    ) -> None:
        """Called for every `node` and/or `attr` on a file's AST."""
        if not hasattr(node, "value"):
            return
        if isinstance(node.value, ast.Call):
            # Search for all of the `default_factory` fields.
            default_factory_value: str | None = None
            for kw in node.value.keywords:
                if kw.arg == "default_factory":
                    # based on the node type, return the proper function name
                    match get_callable_name(kw.value):
                        # `dict` and `list` are common default factory functions
                        case 'dict':
                            default_factory_value = "{}"
                        case 'list':
                            default_factory_value = "[]"

                        case _:
                            # otherwise, see the nodes is in our map for the custom default factory function
                            callable_name = get_callable_name(kw.value)
                            if callable_name in default_factory_map:
                                default_factory_value = pprint.pformat(default_factory_map[callable_name], indent=4, width=80, sort_dicts=False)
                            else:
                                # if not, display as the default
                                default_factory_value = f"{callable_name}()"

            # store the information in the griffe attribute, which is what is passed to the template for rendering
            if "dataclass_ext" not in attr.extra:
                attr.extra["dataclass_ext"] = {}
            attr.extra["dataclass_ext"]["has_default_factory"] = False
            if default_factory_value is not None:
                attr.extra["dataclass_ext"]["has_default_factory"] = True
                attr.extra["dataclass_ext"]["default_factory"] = default_factory_value



from amqtt.broker import BrokerContext
from amqtt.plugins.base import BasePlugin

# intentional import error to test broker response
from pathlib import Pat  # noqa

class MockImportErrorPlugin(BasePlugin):

    def __init__(self, context: BrokerContext) -> None:
        super().__init__(context)



from pathlib import Path

from hypothesis import given, provisional, strategies as st
import yaml

from amqtt import utils
from amqtt.session import Session


@given(st.text())
def test_format_client_message(client_id):
    test_session = Session()
    test_session.client_id = client_id
    client_message = utils.format_client_message(session=test_session)

    assert client_message == f"(client id={client_id})"


@given(provisional.urls(), st.integers())
def test_format_client_message_valid(url, port):
    client_message = utils.format_client_message(address=url, port=port)
    assert client_message == f"(client @={url}:{port})"


def test_format_client_message_unknown():
    client_message = utils.format_client_message()
    assert client_message == "(unknown client)"


def test_client_id():
    client_id = utils.gen_client_id()
    assert isinstance(client_id, str)
    assert client_id.startswith("amqtt/")


def test_read_yaml_config(tmpdir):
    fn = tmpdir / "test.config"
    with Path(fn).open("w") as f:
        yaml.dump({"a": "b"}, f)

    configuration = utils.read_yaml_config(config_file=fn)
    assert configuration == {"a": "b"}



import asyncio
from typing_extensions import Self

from amqtt.adapters import ReaderAdapter
from amqtt.codecs_amqtt import decode_packet_id, decode_string, encode_string, int_to_bytes
from amqtt.errors import AMQTTError, MQTTError
from amqtt.mqtt.packet import PUBLISH, MQTTFixedHeader, MQTTPacket, MQTTPayload, MQTTVariableHeader


class PublishVariableHeader(MQTTVariableHeader):
    __slots__ = ("packet_id", "topic_name")

    def __init__(self, topic_name: str, packet_id: int | None = None) -> None:
        super().__init__()
        if "#" in topic_name or "+" in topic_name:
            msg = "[MQTT-3.3.2-2] Topic name in the PUBLISH Packet MUST NOT contain wildcard characters."
            raise MQTTError(msg)
        self.topic_name = topic_name
        self.packet_id = packet_id

    def __repr__(self) -> str:
        """Return a string representation of the PublishVariableHeader object."""
        return f"{type(self).__name__}(topic={self.topic_name}, packet_id={self.packet_id})"

    def to_bytes(self) -> bytes | bytearray:
        out = bytearray()
        out.extend(encode_string(self.topic_name))
        if self.packet_id is not None:
            out.extend(int_to_bytes(self.packet_id, 2))
        return out

    @classmethod
    async def from_stream(cls, reader: ReaderAdapter | asyncio.StreamReader, fixed_header: MQTTFixedHeader) -> Self:
        topic_name = await decode_string(reader)
        has_qos = (fixed_header.flags >> 1) & 0x03
        packet_id = await decode_packet_id(reader) if has_qos else None
        return cls(topic_name, packet_id)


class PublishPayload(MQTTPayload[MQTTVariableHeader]):
    __slots__ = ("data",)

    def __init__(self, data: bytes | None = None) -> None:
        super().__init__()
        self.data = data

    def to_bytes(
        self,
        fixed_header: MQTTFixedHeader | None = None,
        variable_header: MQTTVariableHeader | None = None,
    ) -> bytes:
        return self.data if self.data is not None else b""

    @classmethod
    async def from_stream(
        cls,
        reader: asyncio.StreamReader | ReaderAdapter,
        fixed_header: MQTTFixedHeader | None,
        variable_header: MQTTVariableHeader | None,
    ) -> Self:
        data = bytearray()
        if fixed_header is None or variable_header is None:
            msg = "Fixed header or variable header cannot be None"
            raise ValueError(msg)

        data_length = fixed_header.remaining_length - variable_header.bytes_length
        length_read = 0
        while length_read < data_length:
            buffer = await reader.read(data_length - length_read)
            data.extend(buffer)
            length_read = len(data)
        return cls(bytes(data))

    def __repr__(self) -> str:
        """Return a string representation of the PublishPayload object."""
        return f"{type(self).__name__}(data={repr(self.data)!r})"


class PublishPacket(MQTTPacket[PublishVariableHeader, PublishPayload, MQTTFixedHeader]):
    VARIABLE_HEADER = PublishVariableHeader
    PAYLOAD = PublishPayload

    DUP_FLAG = 0x08
    RETAIN_FLAG = 0x01
    QOS_FLAG = 0x06

    def __init__(
        self,
        fixed: MQTTFixedHeader | None = None,
        variable_header: PublishVariableHeader | None = None,
        payload: PublishPayload | None = None,
    ) -> None:
        if fixed is None:
            header = MQTTFixedHeader(PUBLISH, 0x00)
        elif fixed.packet_type != PUBLISH:
            msg = f"Invalid fixed packet type {fixed.packet_type} for PublishPacket init"
            raise AMQTTError(msg) from None
        else:
            header = fixed

        super().__init__(header)
        self.variable_header = variable_header
        self.payload = payload

    @classmethod
    def build(cls, topic_name: str, message: bytes, packet_id: int | None, dup_flag: bool, qos: int | None, retain: bool) -> Self:
        v_header = PublishVariableHeader(topic_name, packet_id)
        payload = PublishPayload(message)
        packet = cls(variable_header=v_header, payload=payload)
        packet.dup_flag = dup_flag
        packet.retain_flag = retain
        packet.qos = qos or 0
        return packet

    def set_flags(self, dup_flag: bool = False, qos: int = 0, retain_flag: bool = False) -> None:
        self.dup_flag = dup_flag
        self.retain_flag = retain_flag
        self.qos = qos

    def _set_header_flag(self, val: bool, mask: int) -> None:
        if val:
            self.fixed_header.flags |= mask
        else:
            self.fixed_header.flags &= ~mask

    def _get_header_flag(self, mask: int) -> bool:
        return bool(self.fixed_header.flags & mask)

    @property
    def dup_flag(self) -> bool:
        return self._get_header_flag(self.DUP_FLAG)

    @dup_flag.setter
    def dup_flag(self, val: bool) -> None:
        self._set_header_flag(val, self.DUP_FLAG)

    @property
    def retain_flag(self) -> bool:
        return self._get_header_flag(self.RETAIN_FLAG)

    @retain_flag.setter
    def retain_flag(self, val: bool) -> None:
        self._set_header_flag(val, self.RETAIN_FLAG)

    @property
    def qos(self) -> int | None:
        return (self.fixed_header.flags & self.QOS_FLAG) >> 1

    @qos.setter
    def qos(self, val: int) -> None:
        self.fixed_header.flags &= 0xF9
        self.fixed_header.flags |= val << 1

    @property
    def packet_id(self) -> int | None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        return self.variable_header.packet_id

    @packet_id.setter
    def packet_id(self, val: int) -> None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        self.variable_header.packet_id = val

    @property
    def data(self) -> bytes | None:
        if self.payload is None:
            msg = "Payload header is not set"
            raise ValueError(msg)
        return self.payload.data

    @data.setter
    def data(self, data: bytes) -> None:
        if self.payload is None:
            msg = "Payload header is not set"
            raise ValueError(msg)
        self.payload.data = data

    @property
    def topic_name(self) -> str | None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        return self.variable_header.topic_name

    @topic_name.setter
    def topic_name(self, name: str) -> None:
        if self.variable_header is None:
            msg = "Variable header is not set"
            raise ValueError(msg)
        self.variable_header.topic_name = name



import asyncio
import logging
import signal
import subprocess

from multiprocessing import Process
from pathlib import Path

from typer.testing import CliRunner

from samples.http_server_integration import main as http_server_main
from samples.unix_sockets import app as unix_sockets_app

import pytest

from amqtt.broker import Broker
from amqtt.client import MQTTClient
from samples.broker_acl import config as broker_acl_config
from samples.broker_taboo import config as broker_taboo_config

logger = logging.getLogger(__name__)

@pytest.mark.asyncio
async def test_broker_acl():
    broker_acl_script = Path(__file__).parent.parent / "samples/broker_acl.py"
    process = subprocess.Popen(["python", broker_acl_script], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    # Send the interrupt signal
    await asyncio.sleep(2)
    process.send_signal(signal.SIGINT)
    stdout, stderr = process.communicate()
    logger.debug(stderr.decode("utf-8"))
    assert "Broker closed" in stderr.decode("utf-8")
    assert "ERROR" not in stderr.decode("utf-8")
    assert "Exception" not in stderr.decode("utf-8")


@pytest.mark.asyncio
async def test_broker_simple():
    broker_simple_script = Path(__file__).parent.parent / "samples/broker_simple.py"
    process = subprocess.Popen(["python", broker_simple_script], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    await asyncio.sleep(2)

    # Send the interrupt signal
    process.send_signal(signal.SIGINT)
    stdout, stderr = process.communicate()
    logger.debug(stderr.decode("utf-8"))
    has_broker_closed = "Broker closed" in stderr.decode("utf-8")
    has_loop_stopped = "Broadcast loop stopped by exception" in stderr.decode("utf-8")

    assert has_broker_closed or has_loop_stopped, "Broker didn't close correctly."


@pytest.mark.asyncio
async def test_broker_start():
    broker_start_script = Path(__file__).parent.parent / "samples/broker_start.py"
    process = subprocess.Popen(["python", broker_start_script], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    await asyncio.sleep(2)

    # Send the interrupt signal to stop broker
    process.send_signal(signal.SIGINT)
    stdout, stderr = process.communicate()
    logger.debug(stderr.decode("utf-8"))
    assert "Broker closed" in stderr.decode("utf-8")
    assert "ERROR" not in stderr.decode("utf-8")
    assert "Exception" not in stderr.decode("utf-8")


@pytest.mark.asyncio
async def test_broker_taboo():
    broker_taboo_script = Path(__file__).parent.parent / "samples/broker_taboo.py"
    process = subprocess.Popen(["python", broker_taboo_script], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    await asyncio.sleep(2)

    # Send the interrupt signal to stop broker
    process.send_signal(signal.SIGINT)
    stdout, stderr = process.communicate()
    logger.debug(stderr.decode("utf-8"))
    assert "INFO :: amqtt.broker :: Broker closed" in stderr.decode("utf-8")
    assert "ERROR" not in stderr.decode("utf-8")
    assert "Exception" not in stderr.decode("utf-8")


@pytest.mark.timeout(25)
@pytest.mark.asyncio
async def test_client_keepalive():

    broker = Broker()
    await broker.start()
    await asyncio.sleep(2)

    keep_alive_script = Path(__file__).parent.parent / "samples/client_keepalive.py"
    process = subprocess.Popen(["python", keep_alive_script], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    await asyncio.sleep(1)

    stdout, stderr = process.communicate()
    assert "ERROR" not in stderr.decode("utf-8")
    assert "Exception" not in stderr.decode("utf-8")

    await broker.shutdown()


@pytest.mark.asyncio
async def test_client_publish():
    broker = Broker()
    await broker.start()
    await asyncio.sleep(2)

    client_publish = Path(__file__).parent.parent / "samples/client_publish.py"
    process = subprocess.Popen(["python", client_publish], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    await asyncio.sleep(2)

    stdout, stderr = process.communicate()
    assert "ERROR" not in stderr.decode("utf-8")
    assert "Exception" not in stderr.decode("utf-8")

    await broker.shutdown()


@pytest.fixture
def broker_ssl_config(rsa_keys):
    certfile, keyfile = rsa_keys
    return {
        "listeners": {
            "default": {
                "type": "tcp",
                "bind": "0.0.0.0:8883",
                "ssl": True,
                "certfile": certfile,
                "keyfile": keyfile,
            }
        },
        "auth": {
            "allow-anonymous": True,
            "plugins": ["auth_anonymous"]
        }
    }

@pytest.mark.asyncio
async def test_client_publish_ssl(broker_ssl_config, rsa_keys):
    certfile, _ = rsa_keys
    # generate a self-signed certificate for this test

    # start a secure broker
    broker = Broker(config=broker_ssl_config)
    await broker.start()
    await asyncio.sleep(2)
    # run the sample
    client_publish_ssl_script = Path(__file__).parent.parent / "samples/client_publish_ssl.py"
    process = subprocess.Popen(["python", client_publish_ssl_script, '--cert', certfile], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    await asyncio.sleep(2)
    stdout, stderr = process.communicate()

    assert "ERROR" not in stderr.decode("utf-8")
    assert "Exception" not in stderr.decode("utf-8")

    await broker.shutdown()


@pytest.mark.asyncio
async def test_client_publish_acl():

    broker = Broker()
    await broker.start()
    await asyncio.sleep(2)

    broker_simple_script = Path(__file__).parent.parent / "samples/client_publish_acl.py"
    process = subprocess.Popen(["python", broker_simple_script], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    # Send the interrupt signal
    await asyncio.sleep(2)

    stdout, stderr = process.communicate()
    logger.debug(stderr.decode("utf-8"))
    assert "ERROR" not in stderr.decode("utf-8")
    assert "Exception" not in stderr.decode("utf-8")

    await broker.shutdown()

broker_ws_config = {
    "listeners": {
        "default": {
            "type": "ws",
            "bind": "0.0.0.0:8080",
        }
    },
    "auth": {
        "allow-anonymous": True,
        "plugins": ["auth_anonymous"]
    }
}

@pytest.mark.asyncio
async def test_client_publish_ws():
    # start a secure broker
    broker = Broker(config=broker_ws_config)
    await broker.start()
    await asyncio.sleep(2)
    # run the sample

    client_publish_ssl_script = Path(__file__).parent.parent / "samples/client_publish_ws.py"
    process = subprocess.Popen(["python", client_publish_ssl_script], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    await asyncio.sleep(2)
    stdout, stderr = process.communicate()

    assert "ERROR" not in stderr.decode("utf-8")
    assert "Exception" not in stderr.decode("utf-8")

    await broker.shutdown()


broker_std_config = {
    "listeners": {
        "default": {
            "type": "tcp",
            "bind": "0.0.0.0:1883",
        }
    },
    'sys_interval':2,
    "auth": {
        "allow-anonymous": True,
        "plugins": ["auth_anonymous"]
    }
}


@pytest.mark.asyncio
async def test_client_subscribe():

    # start a standard broker
    broker = Broker(config=broker_std_config)
    await broker.start()
    await asyncio.sleep(1)

    # run the sample
    client_subscribe_script = Path(__file__).parent.parent / "samples/client_subscribe.py"

    process = await asyncio.create_subprocess_shell(
        " ".join(["python", str(client_subscribe_script)]),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await process.communicate()

    assert "ERROR" not in stdout.decode("utf-8")
    assert "Exception" not in stdout.decode("utf-8")
    assert "ERROR" not in stderr.decode("utf-8")
    assert "Exception" not in stderr.decode("utf-8")

    await broker.shutdown()


@pytest.mark.asyncio
async def test_client_subscribe_plugin_acl():
    broker = Broker(config=broker_acl_config)
    await broker.start()

    broker_simple_script = Path(__file__).parent.parent / "samples/client_subscribe_acl.py"
    process = subprocess.Popen(["python", broker_simple_script], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    # Send the interrupt signal
    await asyncio.sleep(2)
    process.send_signal(signal.SIGINT)
    stdout, stderr = process.communicate()
    logger.debug(stderr.decode("utf-8"))
    assert "Subscribed results: [128, 1, 128, 1, 128, 1]" in stderr.decode("utf-8")
    assert "ERROR" not in stderr.decode("utf-8")
    assert "Exception" not in stderr.decode("utf-8")

    await broker.shutdown()


@pytest.mark.asyncio
async def test_client_subscribe_plugin_taboo():
    broker = Broker(config=broker_taboo_config)
    await broker.start()

    broker_simple_script = Path(__file__).parent.parent / "samples/client_subscribe_acl.py"
    process = subprocess.Popen(["python", broker_simple_script], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    # Send the interrupt signal
    await asyncio.sleep(2)
    process.send_signal(signal.SIGINT)
    stdout, stderr = process.communicate()
    logger.debug(stderr.decode("utf-8"))
    assert "Subscribed results: [1, 1, 128, 1, 1, 1]" in stderr.decode("utf-8")
    assert "ERROR" not in stderr.decode("utf-8")
    assert "Exception" not in stderr.decode("utf-8")

    await broker.shutdown()


@pytest.fixture
def external_http_server():
    p = Process(target=http_server_main)
    p.start()
    yield p
    p.terminate()


@pytest.mark.asyncio
async def test_external_http_server(external_http_server):

    await asyncio.sleep(1)
    client = MQTTClient(config={'auto_reconnect': False})
    await client.connect("ws://127.0.0.1:8080/mqtt")
    assert client.session is not None
    await client.publish("my/topic", b'test message')
    await client.disconnect()
    # Send the interrupt signal
    await asyncio.sleep(1)


@pytest.mark.asyncio
async def test_unix_connection():

    unix_socket_script = Path(__file__).parent.parent / "samples/unix_sockets.py"
    broker_process = subprocess.Popen(["python", unix_socket_script, "broker", "-s", "/tmp/mqtt"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    # start the broker
    await asyncio.sleep(1)

    # start the client
    client_process = subprocess.Popen(["python", unix_socket_script, "client", "-s", "/tmp/mqtt"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    await asyncio.sleep(3)

    # stop the client (ctrl-c)
    client_process.send_signal(signal.SIGINT)
    _ = client_process.communicate()

    # stop the broker (ctrl-c)
    broker_process.send_signal(signal.SIGINT)
    broker_stdout, broker_stderr = broker_process.communicate()

    logger.debug(broker_stderr.decode("utf-8"))

    # verify that the broker received client connected/disconnected
    assert "on_broker_client_connected" in broker_stderr.decode("utf-8")
    assert "on_broker_client_disconnected" in broker_stderr.decode("utf-8")



import json
import sqlite3
import tempfile
from pathlib import Path
from unittest.mock import patch, call, ANY

import aiosqlite
import pytest
from jsonschema import validate
from sqlalchemy.ext.asyncio import create_async_engine, async_sessionmaker

from amqtt.broker import BrokerContext, Broker
from amqtt.contrib.shadows import ShadowPlugin
from amqtt.contrib.shadows.models import Shadow, ShadowUpdateError
from amqtt.contrib.shadows.states import StateDocument, State, MetaTimestamp
from amqtt.mqtt.constants import QOS_0
from amqtt.session import IncomingApplicationMessage
from tests.contrib.test_shadows_schema import *


@pytest.fixture
def db_file():
    with tempfile.TemporaryDirectory() as temp_dir:
        with tempfile.NamedTemporaryFile(mode='wb', delete=True) as tmp:
            yield Path(temp_dir) / f"{tmp.name}.db"


@pytest.fixture
def db_connection(db_file):
    test_db_connect = f"sqlite+aiosqlite:///{db_file}"
    yield test_db_connect


@pytest.fixture
@pytest.mark.asyncio
async def db_session_maker(db_connection):
    engine = create_async_engine(f"{db_connection}")
    db_session_maker = async_sessionmaker(engine, expire_on_commit=False)

    yield db_session_maker


@pytest.fixture
@pytest.mark.asyncio
async def shadow_plugin(db_connection):

    cfg = ShadowPlugin.Config(connection=db_connection)
    ctx = BrokerContext(broker=Broker())
    ctx.config = cfg

    shadow_plugin = ShadowPlugin(ctx)
    await shadow_plugin.on_broker_pre_start()
    yield shadow_plugin


@pytest.mark.asyncio
async def test_shadow_find_latest_empty(db_session_maker, shadow_plugin):
    async with db_session_maker() as db_session, db_session.begin():
        shadow = await Shadow.latest_version(session=db_session, device_id='device123', name="myShadowName")
        assert shadow is None


@pytest.mark.asyncio
async def test_shadow_create_new(db_file, db_connection, db_session_maker, shadow_plugin):
    async with db_session_maker() as db_session, db_session.begin():
        shadow = Shadow(device_id='device123', name="myShadowName")
        db_session.add(shadow)
        await db_session.commit()

    async with aiosqlite.connect(db_file) as db_conn:
        db_conn.row_factory = sqlite3.Row  # Set the row_factory

        has_shadow = False
        async with await db_conn.execute("SELECT * FROM shadows_shadow") as cursor:
            for row in await cursor.fetchall():
                assert row['name'] == 'myShadowName'
                assert row['device_id'] == 'device123'
                assert row['state'] == '{}'
                has_shadow = True

        assert has_shadow, "Shadow was not created."

@pytest.mark.asyncio
async def test_shadow_create_find_empty_state(db_connection, db_session_maker, shadow_plugin):
    async with db_session_maker() as db_session, db_session.begin():
        shadow = Shadow(device_id='device123', name="myShadowName")
        db_session.add(shadow)
        await db_session.commit()
        await db_session.flush()

    async with db_session_maker() as db_session, db_session.begin():
        shadow = await Shadow.latest_version(session=db_session, device_id='device123', name="myShadowName")
        assert shadow is not None
        assert shadow.version == 1
        assert shadow.state == StateDocument()


@pytest.mark.asyncio
async def test_shadow_create_find_state_doc(db_connection, db_session_maker, shadow_plugin):
    state_doc = StateDocument(
            state=State(
                desired={'item1': 'value1', 'item2': 'value2'},
                reported={'item3': 'value3', 'item4': 'value4'},
            )
        )
    async with db_session_maker() as db_session, db_session.begin():
        shadow = Shadow(device_id='device123', name="myShadowName")
        shadow.state = state_doc
        db_session.add(shadow)
        await db_session.commit()
        await db_session.flush()

    def new_equal(a, b):
        diff = abs(a.timestamp - b.timestamp)
        return diff <= 2

    async with db_session_maker() as db_session, db_session.begin():
        shadow = await Shadow.latest_version(session=db_session, device_id='device123', name="myShadowName")
        assert shadow is not None
        assert shadow.version == 1
        with patch.object(MetaTimestamp, "__eq__", new=new_equal) as mocked_mqtt_publish:
            assert shadow.state == state_doc

@pytest.mark.asyncio
async def test_shadow_update_state(db_connection, db_session_maker, shadow_plugin):
    state_doc = StateDocument(
            state=State(
                desired={'item1': 'value1', 'item2': 'value2'},
                reported={'item3': 'value3', 'item4': 'value4'},
            )
        )
    async with db_session_maker() as db_session, db_session.begin():
        shadow = Shadow(device_id='device123', name="myShadowName")
        shadow.state = state_doc
        db_session.add(shadow)
        await db_session.commit()
        await db_session.flush()

    async with db_session_maker() as db_session, db_session.begin():
        shadow = await Shadow.latest_version(session=db_session, device_id='device123', name="myShadowName")
        assert shadow is not None
        shadow.state = StateDocument(
            state=State(
                desired={'item5': 'value5', 'item6': 'value6'},
                reported={'item7': 'value7', 'item8': 'value8'},
            )
        )
        with pytest.raises(ShadowUpdateError):
            await db_session.commit()


@pytest.mark.asyncio
async def test_shadow_update_state(db_connection, db_session_maker, shadow_plugin):
    state_doc = StateDocument(
        state=State(
            desired={'item1': 'value1', 'item2': 'value2'},
            reported={'item3': 'value3', 'item4': 'value4'},
        )
    )
    async with db_session_maker() as db_session, db_session.begin():
        shadow = Shadow(device_id='device123', name="myShadowName")
        shadow.state = state_doc
        db_session.add(shadow)
        await db_session.commit()

    async with db_session_maker() as db_session, db_session.begin():
        shadow = await Shadow.latest_version(session=db_session, device_id='device123', name="myShadowName")
        assert shadow is not None
        shadow.state += StateDocument(
            state=State(
                desired={'item1': 'value1a', 'item6': 'value6'}
            )
        )
        await db_session.commit()

    async with db_session_maker() as db_session, db_session.begin():
        shadow_list = await Shadow.all(db_session, "device123", "myShadowName")
        assert len(shadow_list) == 2

    async with db_session_maker() as db_session, db_session.begin():
        shadow = await Shadow.latest_version(session=db_session, device_id='device123', name="myShadowName")
        assert shadow is not None
        assert shadow.version == 2
        assert shadow.state.state.desired == {'item1': 'value1a', 'item2': 'value2', 'item6': 'value6'}
        assert shadow.state.state.reported == {'item3': 'value3', 'item4': 'value4'}


@pytest.mark.asyncio
async def test_shadow_plugin_get_rejected(shadow_plugin):
    """test """

    with patch.object(BrokerContext, 'broadcast_message', return_value=None) as mock_method:
        msg = IncomingApplicationMessage(packet_id=1,
                                             topic='$shadow/myClient123/myShadow/get',
                                             qos=QOS_0,
                                             data=json.dumps({}).encode('utf-8'),
                                             retain=False)
        await shadow_plugin.on_broker_message_received(client_id="myClient123", message=msg)

        mock_method.assert_called()
        topic, message = mock_method.call_args[0]
        assert topic == '$shadow/myClient123/myShadow/get/rejected'
        validate(instance=json.loads(message.decode('utf-8')), schema=get_rejected_schema)

@pytest.mark.asyncio
async def test_shadow_plugin_update_accepted(shadow_plugin):
    with patch.object(BrokerContext, 'broadcast_message', return_value=None) as mock_method:

        update_msg = {
            'state': {
                'desired': {
                    'item1': 'value1',
                    'item2': 'value2'
                }
            }
        }

        validate(instance=update_msg, schema=update_schema)


        msg = IncomingApplicationMessage(packet_id=1,
                                         topic='$shadow/myClient123/myShadow/update',
                                         qos=QOS_0,
                                         data=json.dumps(update_msg).encode('utf-8'),
                                         retain=False)
        await shadow_plugin.on_broker_message_received(client_id="myClient123", message=msg)

        accepted_call = call('$shadow/myClient123/myShadow/update/accepted', ANY)
        document_call = call('$shadow/myClient123/myShadow/update/documents', ANY)
        delta_call = call('$shadow/myClient123/myShadow/update/delta', ANY)
        iota_call = call('$shadow/myClient123/myShadow/update/iota', ANY)

        mock_method.assert_has_calls(
            [
                accepted_call,
                document_call,
                delta_call,
                iota_call,
            ],
            any_order=True,
        )

        for actual in mock_method.call_args_list:
            if actual == accepted_call:
                validate(instance=json.loads(actual.args[1].decode('utf-8')), schema=update_accepted_schema)
            elif actual == document_call:
                validate(instance=json.loads(actual.args[1].decode('utf-8')), schema=update_documents_schema)
            elif actual == delta_call:
                validate(instance=json.loads(actual.args[1].decode('utf-8')), schema=delta_schema)
            elif actual == iota_call:
                validate(instance=json.loads(actual.args[1].decode('utf-8')), schema=delta_schema)
            else:
                assert False, "unknown call made to broadcast"


@pytest.mark.asyncio
async def test_shadow_plugin_get_accepted(shadow_plugin):
    with patch.object(BrokerContext, 'broadcast_message', return_value=None) as mock_method:

        update_msg = {
            'state': {
                'desired': {
                    'item1': 'value1',
                    'item2': 'value2'
                }
            }
        }

        update_msg = IncomingApplicationMessage(packet_id=1,
                                         topic='$shadow/myClient123/myShadow/update',
                                         qos=QOS_0,
                                         data=json.dumps(update_msg).encode('utf-8'),
                                         retain=False)
        await shadow_plugin.on_broker_message_received(client_id="myClient123", message=update_msg)

        mock_method.reset_mock()

        get_msg = IncomingApplicationMessage(packet_id=1,
                                         topic='$shadow/myClient123/myShadow/get',
                                         qos=QOS_0,
                                         data=json.dumps({}).encode('utf-8'),
                                         retain=False)
        await shadow_plugin.on_broker_message_received(client_id="myClient123", message=get_msg)

        get_accepted = call('$shadow/myClient123/myShadow/get/accepted', ANY)

        mock_method.assert_has_calls(
            [get_accepted]
        )

        has_msg = False
        for actual in mock_method.call_args_list:
            if actual == get_accepted:
                validate(instance=json.loads(actual.args[1].decode('utf-8')), schema=get_accepted_schema)
                has_msg = True
        assert has_msg, "could not find the broadcast call for get accepted"



import asyncio
from typing_extensions import Self

from amqtt.adapters import ReaderAdapter
from amqtt.codecs_amqtt import bytes_to_int, int_to_bytes, read_or_raise
from amqtt.errors import AMQTTError, NoDataError
from amqtt.mqtt.packet import SUBACK, MQTTFixedHeader, MQTTPacket, MQTTPayload, MQTTVariableHeader, PacketIdVariableHeader


class SubackPayload(MQTTPayload[MQTTVariableHeader]):
    __slots__ = ("return_codes",)

    RETURN_CODE_00 = 0x00
    RETURN_CODE_01 = 0x01
    RETURN_CODE_02 = 0x02
    RETURN_CODE_80 = 0x80

    def __init__(self, return_codes: list[int] | None = None) -> None:
        super().__init__()
        self.return_codes = return_codes or []

    def __repr__(self) -> str:
        """Return a string representation of the SubackPayload object."""
        return f"{type(self).__name__}(return_codes={self.return_codes!r})"

    def to_bytes(
        self,
        fixed_header: MQTTFixedHeader | None = None,
        variable_header: MQTTVariableHeader | None = None,
    ) -> bytes:
        out = b""
        for return_code in self.return_codes:
            out += int_to_bytes(return_code, 1)
        return out

    @classmethod
    async def from_stream(
        cls,
        reader: asyncio.StreamReader | ReaderAdapter,
        fixed_header: MQTTFixedHeader | None,
        variable_header: MQTTVariableHeader | None,
    ) -> Self:
        return_codes = []
        if fixed_header is None or variable_header is None:
            msg = "Fixed header or variable header cannot be None"
            raise AMQTTError(msg)

        bytes_to_read = fixed_header.remaining_length - variable_header.bytes_length
        for _ in range(bytes_to_read):
            try:
                return_code_byte = await read_or_raise(reader, 1)
                return_code = bytes_to_int(return_code_byte)
                return_codes.append(return_code)
            except NoDataError:
                break
        return cls(return_codes)


class SubackPacket(MQTTPacket[PacketIdVariableHeader, SubackPayload, MQTTFixedHeader]):
    VARIABLE_HEADER = PacketIdVariableHeader
    PAYLOAD = SubackPayload

    def __init__(
        self,
        fixed: MQTTFixedHeader | None = None,
        variable_header: PacketIdVariableHeader | None = None,
        payload: SubackPayload | None = None,
    ) -> None:
        if fixed is None:
            header = MQTTFixedHeader(SUBACK, 0x00)
        else:
            if fixed.packet_type is not SUBACK:
                msg = f"Invalid fixed packet type {fixed.packet_type} for SubackPacket init"
                raise AMQTTError(msg)
            header = fixed

        super().__init__(header)
        self.variable_header = variable_header
        self.payload = payload

    @classmethod
    def build(cls, packet_id: int, return_codes: list[int]) -> Self:
        variable_header = cls.VARIABLE_HEADER(packet_id)
        payload = cls.PAYLOAD(return_codes)
        return cls(variable_header=variable_header, payload=payload)



from dataclasses import dataclass, is_dataclass
from typing import Any, Generic, TypeVar, cast

from amqtt.contexts import Action, BaseContext, BrokerConfig
from amqtt.session import Session

C = TypeVar("C", bound=BaseContext)


class BasePlugin(Generic[C]):
    """The base from which all plugins should inherit.

    Type Parameters
    ---------------
    C:
        A BaseContext: either BrokerContext or ClientContext, depending on plugin usage

    Attributes
    ----------
    context (C):
        Information about the environment in which this plugin is executed. Modifying
        the broker or client state should happen through methods available here.

    config (self.Config):
        An instance of the Config dataclass defined by the plugin (or an empty dataclass, if not
        defined). If using entrypoint- or mixed-style configuration, use `_get_config_option()`
        to access the variable.

    """

    def __init__(self, context: C) -> None:
        self.context: C = context
        # since the PluginManager will hydrate the config from a plugin's `Config` class, this is a safe cast
        self.config = cast("self.Config", context.config)  # type: ignore[name-defined]

    # Deprecated: included to support entrypoint-style configs. Replaced by dataclass Config class.
    def _get_config_section(self, name: str) -> dict[str, Any] | None:

        if not self.context.config or not hasattr(self.context.config, "get") or not self.context.config.get(name, None):
            return None

        section_config: int | dict[str, Any] | None = self.context.config.get(name, None)
        # mypy has difficulty excluding int from `config`'s type, unless there's an explicit check
        if isinstance(section_config, int):
            return None
        return section_config

    # Deprecated : supports entrypoint-style configs as well as dataclass configuration.
    def _get_config_option(self, option_name: str, default: Any = None) -> Any:
        if not self.context.config:
            return default

        if is_dataclass(self.context.config):
            # overloaded context.config for BasePlugin `Config` class, so ignoring static type check
            return getattr(self.context.config, option_name.replace("-", "_"), default)
        if option_name in self.context.config:
            return self.context.config[option_name]
        return default

    @dataclass
    class Config:
        """Override to define the configuration and defaults for plugin."""

    async def close(self) -> None:
        """Override if plugin needs to clean up resources upon shutdown."""


class BaseTopicPlugin(BasePlugin[BaseContext]):
    """Base class for topic plugins."""

    def __init__(self, context: BaseContext) -> None:
        super().__init__(context)

        self.topic_config: dict[str, Any] | None = self._get_config_section("topic-check")
        if not bool(self.topic_config) and not is_dataclass(self.context.config):
            self.context.logger.warning("'topic-check' section not found in context configuration")

    def _get_config_option(self, option_name: str, default: Any = None) -> Any:
        if not self.context.config:
            return default

        # overloaded context.config with either BrokerConfig or plugin's Config
        if is_dataclass(self.context.config) and not isinstance(self.context.config, BrokerConfig):
            # overloaded context.config for BasePlugin `Config` class, so ignoring static type check
            return getattr(self.context.config, option_name.replace("-", "_"), default)
        if self.topic_config and option_name in self.topic_config:
            return self.topic_config[option_name]
        return default

    async def topic_filtering(
        self, *, session: Session | None = None, topic: str | None = None, action: Action | None = None
    ) -> bool | None:
        """Logic for filtering out topics.

        Args:
            session: amqtt.session.Session
            topic: str
            action: amqtt.broker.Action

        Returns:
            bool: `True` if topic is allowed, `False` otherwise. `None` if it can't be determined

        """
        return bool(self.topic_config) or is_dataclass(self.context.config)


class BaseAuthPlugin(BasePlugin[BaseContext]):
    """Base class for authentication plugins."""

    def _get_config_option(self, option_name: str, default: Any = None) -> Any:
        if not self.context.config:
            return default

        if is_dataclass(self.context.config) and not isinstance(self.context.config, BrokerConfig):
            # overloaded context.config for BasePlugin `Config` class, so ignoring static type check
            return getattr(self.context.config, option_name.replace("-", "_"), default)
        if self.auth_config and option_name in self.auth_config:
            return self.auth_config[option_name]
        return default

    def __init__(self, context: BaseContext) -> None:
        super().__init__(context)

        self.auth_config: dict[str, Any] | None = self._get_config_section("auth")
        if not bool(self.auth_config) and not is_dataclass(self.context.config):
            # auth config section not found and Config dataclass not provided
            self.context.logger.warning("'auth' section not found in context configuration")

    async def authenticate(self, *, session: Session) -> bool | None:
        """Logic for session authentication.

        Args:
            session: amqtt.session.Session

        Returns:
            - `True` if user is authentication succeed, `False` if user authentication fails
            - `None` if authentication can't be achieved (then plugin result is then ignored)

        """
        return bool(self.auth_config) or is_dataclass(self.context.config)



from dataclasses import dataclass
import logging
from typing import ClassVar

import ldap

from amqtt.broker import BrokerContext
from amqtt.contexts import Action
from amqtt.errors import PluginInitError
from amqtt.plugins import TopicMatcher
from amqtt.plugins.base import BaseAuthPlugin, BasePlugin, BaseTopicPlugin
from amqtt.session import Session

logger = logging.getLogger(__name__)


@dataclass
class LdapConfig:
    """Configuration for the LDAP Plugins."""

    server: str
    """uri formatted server location. e.g `ldap://localhost:389`"""
    base_dn: str
    """distinguished name (dn) of the ldap server. e.g. `dc=amqtt,dc=io`"""
    user_attribute: str
    """attribute in ldap entry to match the username against"""
    bind_dn: str
    """distinguished name (dn) of known, preferably read-only, user. e.g. `cn=admin,dc=amqtt,dc=io`"""
    bind_password: str
    """password for known, preferably read-only, user"""


class AuthLdapPlugin(BasePlugin[BrokerContext]):

    def __init__(self, context: BrokerContext) -> None:
        super().__init__(context)

        self.conn = ldap.initialize(self.config.server)
        self.conn.protocol_version = ldap.VERSION3  # pylint: disable=E1101
        try:
            self.conn.simple_bind_s(self.config.bind_dn, self.config.bind_password)
        except ldap.INVALID_CREDENTIALS as e:  # pylint: disable=E1101
            raise PluginInitError(self.__class__) from e


class UserAuthLdapPlugin(AuthLdapPlugin, BaseAuthPlugin):
    """Plugin to authenticate a user with an LDAP directory server."""

    async def authenticate(self, *, session: Session) -> bool | None:

        # use our initial creds to see if the user exists
        search_filter = f"({self.config.user_attribute}={session.username})"
        result = self.conn.search_s(self.config.base_dn, ldap.SCOPE_SUBTREE, search_filter, ["dn"])  # pylint: disable=E1101
        if not result:
            logger.debug(f"user not found: {session.username}")
            return False

        try:
            # `search_s` responds with list of tuples: (dn, entry); first in list is our match
            user_dn = result[0][0]
        except IndexError:
            return False

        try:
            user_conn = ldap.initialize(self.config.server)
            user_conn.simple_bind_s(user_dn, session.password)
        except ldap.INVALID_CREDENTIALS:  # pylint: disable=E1101
            logger.debug(f"invalid credentials for '{session.username}'")
            return False
        except ldap.LDAPError as e:  # pylint: disable=E1101
            logger.debug(f"LDAP error during user bind: {e}")
            return False

        return True

    @dataclass
    class Config(LdapConfig):
        """Configuration for the User Auth LDAP Plugin."""


class TopicAuthLdapPlugin(AuthLdapPlugin, BaseTopicPlugin):
    """Plugin to authenticate a user with an LDAP directory server."""

    _action_attr_map: ClassVar = {
        Action.PUBLISH: "publish_attribute",
        Action.SUBSCRIBE: "subscribe_attribute",
        Action.RECEIVE: "receive_attribute"
    }

    def __init__(self, context: BrokerContext) -> None:
        super().__init__(context)

        self.topic_matcher = TopicMatcher()

    async def topic_filtering(
        self, *, session: Session | None = None, topic: str | None = None, action: Action | None = None
    ) -> bool | None:

        # if not provided needed criteria, can't properly evaluate topic filtering
        if not session or not action or not topic:
            return None

        search_filter = f"({self.config.user_attribute}={session.username})"
        attrs = [
            "cn",
            self.config.publish_attribute,
            self.config.subscribe_attribute,
            self.config.receive_attribute
        ]
        results = self.conn.search_s(self.config.base_dn, ldap.SCOPE_SUBTREE, search_filter, attrs)  # pylint: disable=E1101

        if not results:
            logger.debug(f"user not found: {session.username}")
            return False

        if len(results) > 1:
            found_users = [dn for dn, _ in results]
            logger.debug(f"multiple users found: {', '.join(found_users)}")
            return False

        dn, entry = results[0]

        ldap_attribute = getattr(self.config, self._action_attr_map[action])
        topic_filters = [t.decode("utf-8") for t in entry.get(ldap_attribute, [])]
        logger.debug(f"DN: {dn} - {ldap_attribute}={topic_filters}")

        return self.topic_matcher.are_topics_allowed(topic, topic_filters)

    @dataclass
    class Config(LdapConfig):
        """Configuration for the LDAPAuthPlugin."""

        publish_attribute: str
        """LDAP attribute which contains a list of permissible publish topics."""
        subscribe_attribute: str
        """LDAP attribute which contains a list of permissible subscribe topics."""
        receive_attribute: str
        """LDAP attribute which contains a list of permissible receive topics."""



import asyncio
import inspect
import logging
from functools import partial
from logging import getLogger
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, Coroutine

import pytest

import amqtt.plugins
from amqtt.broker import Broker, BrokerContext
from amqtt.client import MQTTClient
from amqtt.errors import PluginInitError, PluginImportError
from amqtt.events import MQTTEvents, BrokerEvents
from amqtt.mqtt.constants import QOS_0, QOS_1
from amqtt.plugins.base import BasePlugin
from amqtt.contexts import BaseContext
from amqtt.contrib.persistence import RetainedMessage

_INVALID_METHOD: str = "invalid_foo"
_PLUGIN: str = "Plugin"

logger = logging.getLogger(__name__)

class _TestContext(BaseContext):
    def __init__(self) -> None:
        super().__init__()
        self.config: dict[str, Any] = {"auth": {}}
        self.logger = getLogger(__name__)


def _verify_module(module: ModuleType, plugin_module_name: str) -> None:
    if not module.__name__.startswith(plugin_module_name):
        return

    for name, clazz in inspect.getmembers(module, inspect.isclass):
        if not name.endswith(_PLUGIN) or name == _PLUGIN:
            continue

        obj = clazz(_TestContext())
        with pytest.raises(
            AttributeError,
            match=f"'{name}' object has no attribute '{_INVALID_METHOD}'",
        ):
            getattr(obj, _INVALID_METHOD)
        assert hasattr(obj, _INVALID_METHOD) is False

    for _, obj in inspect.getmembers(module, inspect.ismodule):
        _verify_module(obj, plugin_module_name)


def removesuffix(self: str, suffix: str) -> str:
    """Compatibility for Python versions prior to 3.9."""
    if suffix and self.endswith(suffix):
        return self[: -len(suffix)]
    return self[:]


def test_plugins_correct_has_attr() -> None:
    """Test plugins to ensure they correctly handle the 'has_attr' check."""
    module = amqtt.plugins
    for file in Path(module.__file__).parent.rglob("*.py"):
        if not Path(file).is_file():
            continue

        name = file.as_posix().replace("/", ".")
        name = name[name.find(module.__name__) : -3]
        name = removesuffix(name, ".__init__")

        __import__(name)

    _verify_module(module, module.__name__)


class MockInitErrorPlugin(BasePlugin):

    def __init__(self, context: BrokerContext) -> None:
        super().__init__(context)
        raise KeyError


@pytest.mark.asyncio
async def test_plugin_exception_while_init() -> None:

    config = {
        "listeners": {
            "default": {"type": "tcp", "bind": "127.0.0.1:1883", "max_connections": 10},
        },
        'sys_interval': 1,
        'plugins':{
            'tests.plugins.test_plugins.MockInitErrorPlugin':{}
        }
    }

    with pytest.raises(PluginInitError):
        _ = Broker(plugin_namespace='tests.mock_plugins', config=config)


@pytest.mark.asyncio
async def test_plugin_exception_while_loading() -> None:

    config = {
        "listeners": {
            "default": {"type": "tcp", "bind": "127.0.0.1:1883", "max_connections": 10},
        },
        'sys_interval': 1,
        'plugins':{
            'tests.plugins.mock_plugins.MockImportErrorPlugin':{}
        }
    }

    with pytest.raises(PluginImportError):
        _ = Broker(plugin_namespace='tests.mock_plugins', config=config)


class AllEventsPlugin(BasePlugin[BaseContext]):
    """A plugin to verify all events get sent to plugins."""
    def __init__(self, context: BaseContext) -> None:
        super().__init__(context)

        self.test_flags = { events:False for events in list(MQTTEvents) + list(BrokerEvents)}

    async def call_method(self, event_name: str, **kwargs: Any) -> None:
        assert event_name in self.test_flags
        self.test_flags[event_name] = True

    def __getattr__(self, name: str) -> Callable[..., Coroutine[Any, Any, None]]:
        """Dynamically handle calls to methods starting with 'on_'."""

        if name.startswith("on_"):
            event_name = name.replace('on_', '')
            return partial(self.call_method, event_name)

        if name not in ('authenticate', 'topic_filtering'):
            pytest.fail(f'unexpected method called: {name}')


@pytest.mark.asyncio
async def test_all_plugin_events():

    config = {
        "listeners": {
            "default": {"type": "tcp", "bind": "127.0.0.1:1883", "max_connections": 10},
        },
        'sys_interval': 1,
        'plugins':{
            'amqtt.plugins.authentication.AnonymousAuthPlugin': {},
            'tests.plugins.test_plugins.AllEventsPlugin': {}
        }
    }

    broker = Broker(plugin_namespace='tests.mock_plugins', config=config)

    await broker.start()
    await asyncio.sleep(2)

    # make sure all expected events get triggered
    client = MQTTClient()
    await client.connect("mqtt://127.0.0.1:1883/")
    await client.subscribe([('my/test/topic', QOS_0),])
    await client.publish('test/topic', b'my test message', retain=True)
    await client.unsubscribe(['my/test/topic',])
    await client.disconnect()
    await asyncio.sleep(1)

    # get the plugin so it doesn't get gc on shutdown
    test_plugin = broker.plugins_manager.get_plugin('AllEventsPlugin')
    await broker.shutdown()
    await asyncio.sleep(1)

    assert all(test_plugin.test_flags.values()), f'event not received: {[event for event, value in test_plugin.test_flags.items() if not value]}'


class RetainedMessageEventPlugin(BasePlugin[BrokerContext]):
    """A plugin to verify all events get sent to plugins."""
    def __init__(self, context: BaseContext) -> None:
        super().__init__(context)
        self.topic_retained_message_flag = False
        self.session_retained_message_flag = False
        self.topic_clear_retained_message_flag = False

    async def on_broker_retained_message(self, *, client_id: str | None, retained_message: RetainedMessage) -> None:
        """retaining message event handler."""
        if client_id:
            session = self.context.get_session(client_id)
            assert session.transitions.state != "connected"
            logger.debug("retained message event fired for offline client")
            self.session_retained_message_flag = True
        else:
            if not retained_message.data:
                logger.debug("retained message event fired for clearing a topic")
                self.topic_clear_retained_message_flag = True
            else:
                logger.debug("retained message event fired for setting a topic")
                self.topic_retained_message_flag = True


@pytest.mark.asyncio
async def test_retained_message_plugin_event():

    config = {
        "listeners": {
            "default": {"type": "tcp", "bind": "127.0.0.1:1883", "max_connections": 10},
        },
        'sys_interval': 1,
        'plugins':[{'amqtt.plugins.authentication.AnonymousAuthPlugin': {'allow_anonymous': False}},
                   {'tests.plugins.test_plugins.RetainedMessageEventPlugin': {}}]
        }

    broker = Broker(plugin_namespace='tests.mock_plugins', config=config)

    await broker.start()
    await asyncio.sleep(0.1)

    # make sure all expected events get triggered
    client1 = MQTTClient(config={'auto_reconnect': False})
    await client1.connect("mqtt://myUsername@127.0.0.1:1883/", cleansession=False)
    await client1.subscribe([('test/topic', QOS_1),])
    await client1.publish('test/retained', b'message should be retained for test/retained', retain=True)
    await asyncio.sleep(0.1)
    await client1.disconnect()

    client2 = MQTTClient(config={'auto_reconnect': False})
    await client2.connect("mqtt://myOtherUsername@127.0.0.1:1883/", cleansession=True)
    await client2.publish('test/topic', b'message should be retained for myUsername since subscription was qos > 0')
    await client2.publish('test/retained', b'', retain=True)  # should clear previously retained message
    await asyncio.sleep(0.1)
    await client2.disconnect()
    await asyncio.sleep(0.1)

    # get the plugin so it doesn't get gc on shutdown
    test_plugin = broker.plugins_manager.get_plugin('RetainedMessageEventPlugin')
    await broker.shutdown()
    await asyncio.sleep(0.1)

    assert test_plugin.topic_retained_message_flag, "message to topic wasn't retained"
    assert test_plugin.session_retained_message_flag, "message to disconnected client wasn't retained"
    assert test_plugin.topic_clear_retained_message_flag, "message to retained topic wasn't cleared"



import asyncio
import logging
from dataclasses import dataclass, field
from typing import Any

import pytest
import yaml

from amqtt.broker import Broker
from yaml import CLoader as Loader
from dacite import from_dict, Config, UnexpectedDataError

from amqtt.client import MQTTClient
from amqtt.errors import PluginLoadError, ConnectError, PluginCoroError
from amqtt.mqtt.constants import QOS_0

logger = logging.getLogger(__name__)

plugin_config = """---
listeners:
  default:
    type: tcp
    bind: 0.0.0.0:1883
plugins:
   - tests.plugins.mocks.TestSimplePlugin:
   - tests.plugins.mocks.TestConfigPlugin:
       option1: 1
       option2: bar
"""


plugin_invalid_config_one = """---
listeners:
  default:
    type: tcp
    bind: 0.0.0.0:1883
plugins:
   - tests.plugins.mocks.TestSimplePlugin:
       option1: 1
       option2: bar
"""

plugin_invalid_config_two = """---
listeners:
  default:
    type: tcp
    bind: 0.0.0.0:1883
plugins:
   - tests.plugins.mocks.TestConfigPlugin:
"""

plugin_coro_error_config = """---
listeners:
  default:
    type: tcp
    bind: 0.0.0.0:1883
plugins:
   - tests.plugins.mocks.TestCoroErrorPlugin:
"""

plugin_config_auth = """---
listeners:
  default:
    type: tcp
    bind: 0.0.0.0:1883
plugins:
   - tests.plugins.mocks.TestAuthPlugin:
"""

plugin_config_no_auth = """---
listeners:
  default:
    type: tcp
    bind: 0.0.0.0:1883
plugins:
   - tests.plugins.mocks.TestNoAuthPlugin:
"""


plugin_config_topic = """---
listeners:
  default:
    type: tcp
    bind: 0.0.0.0:1883
plugins:
   - amqtt.plugins.authentication.AnonymousAuthPlugin
   - tests.plugins.mocks.TestAllowTopicPlugin:
"""


plugin_config_topic_block = """---
listeners:
  default:
    type: tcp
    bind: 0.0.0.0:1883
plugins:
   - amqtt.plugins.authentication.AnonymousAuthPlugin
   - tests.plugins.mocks.TestBlockTopicPlugin:
"""



@pytest.mark.asyncio
async def test_plugin_config_extra_fields():

    cfg: dict[str, Any] = yaml.load(plugin_invalid_config_one, Loader=Loader)

    with pytest.raises(PluginLoadError):
        _ = Broker(config=cfg)


@pytest.mark.asyncio
async def test_plugin_config_missing_fields():
    cfg: dict[str, Any] = yaml.load(plugin_invalid_config_one, Loader=Loader)

    with pytest.raises(PluginLoadError):
        _ = Broker(config=cfg)


@pytest.mark.asyncio
async def test_alternate_plugin_load():

    cfg: dict[str, Any] = yaml.load(plugin_config, Loader=Loader)

    broker = Broker(config=cfg)
    await broker.start()
    await broker.shutdown()


@pytest.mark.asyncio
async def test_coro_error_plugin_load():

    cfg: dict[str, Any] = yaml.load(plugin_coro_error_config, Loader=Loader)

    with pytest.raises(PluginCoroError):
        _ = Broker(config=cfg)


@pytest.mark.asyncio
async def test_auth_plugin_load():
    cfg: dict[str, Any] = yaml.load(plugin_config_auth, Loader=Loader)
    broker = Broker(config=cfg)
    await broker.start()
    await asyncio.sleep(0.5)

    client1 = MQTTClient()
    await client1.connect()
    await client1.publish('my/topic', b'my message')
    await client1.disconnect()

    await asyncio.sleep(0.5)
    await broker.shutdown()


@pytest.mark.asyncio
async def test_no_auth_plugin_load():
    cfg: dict[str, Any] = yaml.load(plugin_config_no_auth, Loader=Loader)
    broker = Broker(config=cfg)
    await broker.start()
    await asyncio.sleep(0.5)

    client1 = MQTTClient(config={'auto_reconnect': False})
    with pytest.raises(ConnectError):
        await client1.connect()

    await asyncio.sleep(0.5)
    await broker.shutdown()


@pytest.mark.asyncio
async def test_allow_topic_plugin_load():
    cfg: dict[str, Any] = yaml.load(plugin_config_topic, Loader=Loader)
    broker = Broker(config=cfg)
    await broker.start()
    await asyncio.sleep(0.5)

    client2 = MQTTClient(config={'auto_reconnect': False})
    await client2.connect()
    await client2.subscribe([
        ('my/topic', QOS_0)
    ])

    client1 = MQTTClient(config={'auto_reconnect': True})
    await client1.connect()
    await client1.publish('my/topic', b'my message')

    message = await client2.deliver_message(timeout_duration=1)
    assert message.topic == 'my/topic'
    assert message.data == b'my message'

    await client2.disconnect()
    await client1.disconnect()

    await broker.shutdown()


@pytest.mark.asyncio
async def test_block_topic_plugin_load():
    cfg: dict[str, Any] = yaml.load(plugin_config_topic_block, Loader=Loader)
    broker = Broker(config=cfg)
    await broker.start()
    await asyncio.sleep(0.5)

    client2 = MQTTClient(config={'auto_reconnect': False})
    await client2.connect()
    await client2.subscribe([
        ('my/topic', QOS_0)
    ])

    client1 = MQTTClient(config={'auto_reconnect': True})
    await client1.connect()
    await client1.publish('my/topic', b'my message')

    with pytest.raises(asyncio.TimeoutError):
        message = await client2.deliver_message(timeout_duration=1)
        logger.debug(f"msg received: {message.topic} >> {message.data}")

    await client2.disconnect()
    await client1.disconnect()

    await broker.shutdown()

plugin_yaml_list_config_one = """---
listeners:
  default:
    type: tcp
    bind: 0.0.0.0:1883
plugins:
   - tests.plugins.mocks.TestSimplePlugin:
   - tests.plugins.mocks.TestConfigPlugin:
       option1: 1
       option2: bar
       option3: 3
"""

plugin_yaml_list_config_two = """---
listeners:
  default:
    type: tcp
    bind: 0.0.0.0:1883
plugins:
   - tests.plugins.mocks.TestSimplePlugin
   - tests.plugins.mocks.TestConfigPlugin:
       option1: 1
       option2: bar
       option3: 3
"""

plugin_yaml_dict_config = """---
listeners:
  default:
    type: tcp
    bind: 0.0.0.0:1883
plugins:
   tests.plugins.mocks.TestSimplePlugin:
   tests.plugins.mocks.TestConfigPlugin:
    option1: 1
    option2: bar
    option3: 3
"""

plugin_empty_dict_config = {
    'listeners': {'default': {'type': 'tcp', 'bind': '127.0.0.1'}},
    'plugins': {
        'tests.plugins.mocks.TestSimplePlugin': {},
    }
}

plugin_dict_option_config = {
    'listeners': {'default': {'type': 'tcp', 'bind': '127.0.0.1'}},
    'plugins': {
        'tests.plugins.mocks.TestConfigPlugin': {'option1': 1, 'option2': 'bar', 'option3': 3}
    }
}

@pytest.mark.asyncio
async def test_plugin_yaml_list_config():
    cfg: dict[str, Any] = yaml.load(plugin_yaml_list_config_one, Loader=Loader)
    broker = Broker(config=cfg)

    await asyncio.sleep(0.5)
    plugin = broker.plugins_manager.get_plugin('TestConfigPlugin')
    assert getattr(plugin.context.config, 'option1', None) == 1
    assert getattr(plugin.context.config, 'option3', None) == 3

    cfg: dict[str, Any] = yaml.load(plugin_yaml_list_config_two, Loader=Loader)
    broker = Broker(config=cfg)

    await asyncio.sleep(0.5)
    plugin = broker.plugins_manager.get_plugin('TestConfigPlugin')
    assert getattr(plugin.context.config, 'option1', None) == 1
    assert getattr(plugin.context.config, 'option3', None) == 3


@pytest.mark.asyncio
async def test_plugin_yaml_dict_config():
    cfg: dict[str, Any] = yaml.load(plugin_yaml_dict_config, Loader=Loader)
    broker = Broker(config=cfg)

    await asyncio.sleep(0.5)
    assert broker.plugins_manager.get_plugin('TestSimplePlugin') is not None


@pytest.mark.asyncio
async def test_plugin_empty_dict_config():
    broker = Broker(config=plugin_empty_dict_config)

    await asyncio.sleep(0.5)
    assert broker.plugins_manager.get_plugin('TestSimplePlugin') is not None


@pytest.mark.asyncio
async def test_plugin_option_dict_config():
    broker = Broker(config=plugin_dict_option_config)

    await asyncio.sleep(0.5)
    plugin = broker.plugins_manager.get_plugin('TestConfigPlugin')
    assert getattr(plugin.context.config, 'option1', None) == 1
    assert getattr(plugin.context.config, 'option3', None) == 3
